"""
MeshInfo
==========
"""
from ansys.dpf.core import server as server_module
from ansys.dpf.core.generic_data_container import GenericDataContainer


class MeshInfo:
    """Hold the information relative to a mesh region.

    This class describes the available mesh information.

    Parameters
    ----------
     server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.
     generic_data_container : ansys.dpf.core.generic_data_container, optional
        Generic data container that is wrapped into the mesh info.
     mesh_info : optional
        Hold the information of the mesh region into a generic data container.

    Examples
    --------
    Explore the mesh info from the model

    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> fluent = examples.fluid_axial_model()
    >>> model = dpf.Model(fluent)
    >>> mesh_info = model.metadata.mesh_info

    Notes
    -----
    Class available with server's version starting at 7.0 (2024 R1 pre0).
    """

    def __init__(
        self,
        generic_data_container=None,
        mesh_info=None,
        server=None,
    ):
        """Initialize with a MeshInfo message"""
        # ############################
        # step 1: get server

        if generic_data_container is None and mesh_info is None:
            self._server = server_module.get_or_create_server(server)
            self._generic_data_container = GenericDataContainer(server=self._server)
        elif generic_data_container is not None and mesh_info is None:
            self._server = generic_data_container._server
            self._generic_data_container = generic_data_container
        elif generic_data_container is None and MeshInfo is not None:
            self._server = mesh_info.generic_data_container._server
            self._generic_data_container = mesh_info.generic_data_container
        else:
            raise ValueError(
                "Arguments generic_data_container and mesh_info are mutually exclusive."
            )

    def __str__(self):
        txt = "DPF MeshInfo\n"
        txt += "-" * 30 + "\n"
        txt += "with properties:\n"
        txt += "\n".join(
            "  {:20s}\t{:s}".format(k, v)
            for k, v in self._generic_data_container.get_property_description().items()
        )
        return txt

    @property
    def generic_data_container(self) -> GenericDataContainer:
        """GenericDataContainer wrapped into the MeshInfo
        that contains all the relative information of the derived class.

        Returns
        -------
        :class:`ansys.dpf.core.generic_data_container.GenericDataContainer`

        """

        return self._generic_data_container

    @generic_data_container.setter
    def generic_data_container(self, value: GenericDataContainer):
        """GenericDataContainer wrapped into the MeshInfo
        that contains all the relative information of the derived class.
        """

        if not isinstance(value, GenericDataContainer):
            raise ValueError("Input value must be a GenericDataContainer.")
        self._generic_data_container = value
        self._server = self._generic_data_container._server

    def deep_copy(self, server=None):
        """Create a deep copy of the mesh_info's data on a given server.

        This method is useful for passing data from one server instance to another.

        Parameters
        ----------
        server : ansys.dpf.core.server, optional
            Server with the channel connected to the remote or local instance.
            The default is ``None``, in which case an attempt is made to use the
            global server.

        Returns
        -------
        mesh_info_copy : MeshInfo
        """
        mesh_info = MeshInfo(server=server)
        mesh_info.generic_data_container = self._generic_data_container.deep_copy(server)
        return mesh_info

    def get_property(self, property_name):
        """Get property with given name.

        Parameters
        ----------
        property_name : str
            Property name.

        Returns
        -------
        Property object instance.
        """
        return self.generic_data_container.get_property(property_name)

    def set_property(self, property_name, prop):
        """Register given property with the given name.

        Parameters
        ----------
        property_name : str
            Property name.
        prop : Int, String, Float, Field, StringField, GenericDataContainer, Scoping
            object instance.
        """

        return self.generic_data_container.set_property(property_name, prop)

    @property
    def number_nodes(self):
        """
        Returns
        -------
        number_nodes : int
            Number of nodes of the mesh.
        """

        return self.generic_data_container.get_property("num_nodes")

    @property
    def number_faces(self):
        """
        Returns
        -------
        number_faces : int
            Number of faces of the mesh.
        """
        if "num_faces" in self._generic_data_container.get_property_description():
            return self.generic_data_container.get_property("num_faces")
        else:
            return 0

    @property
    def number_elements(self):
        """
        Returns
        -------
        number_elements : int
            Number of elements of the mesh.
        """

        if "num_cells" in self._generic_data_container.get_property_description():
            return self.generic_data_container.get_property("num_cells")
        else:
            return self.generic_data_container.get_property("num_elements")

    @property
    def splittable_by(self):
        """
        Returns
        -------
        splittable by which entity : StringField
            Name of the properties according to which the mesh can be split by.
        """

        if "splittable_by" in self._generic_data_container.get_property_description():
            return self.generic_data_container.get_property("splittable_by")
        else:
            return None

    @property
    def available_elem_types(self):
        """
        Returns
        -------
        available element types : Scoping
            element type available for the mesh.
        """

        if "available_elem_types" in self._generic_data_container.get_property_description():
            return self.generic_data_container.get_property("available_elem_types")
        else:
            return None

    @property
    def part_names(self):
        """
        Returns
        -------
        part_names : StringField
            part names of the mesh (if it can be split by parts)
        """

        if "part_names" in self._generic_data_container.get_property_description():
            return self.generic_data_container.get_property("part_names")
        else:
            return None

    @property
    def part_scoping(self):
        """
        Returns
        -------
        part_scoping : Scoping
            part Scoping of the mesh (if it can be split by parts)
        """

        if "part_scoping" in self._generic_data_container.get_property_description():
            return self.generic_data_container.get_property("part_scoping")
        else:
            return None

    @property
    def body_names(self):
        """
        Returns
        -------
        body_names : StringField
            body names of the mesh (if it can be split by bodies)
        """

        if "body_names" in self._generic_data_container.get_property_description():
            return self.generic_data_container.get_property("body_names")
        else:
            return None

    @property
    def body_scoping(self):
        """
        Returns
        -------
        body_scoping : Scoping
            body Scoping of the mesh (if it can be split by bodies)
        """

        if "body_scoping" in self._generic_data_container.get_property_description():
            return self.generic_data_container.get_property("body_scoping")
        else:
            return None

    @property
    def zone_names(self):
        """
        Returns
        -------
        zone_names : StringField
            zone_names of the mesh (if it can be split by zones)
        """

        if "zone_names" in self._generic_data_container.get_property_description():
            return self.generic_data_container.get_property("zone_names")
        else:
            return None

    @property
    def zone_scoping(self):
        """
        Returns
        -------
        zone_scoping : Scoping
            zone Scoping of the mesh (if it can be split by zone)
        """

        if "zone_scoping" in self._generic_data_container.get_property_description():
            return self.generic_data_container.get_property("zone_scoping")
        else:
            return None

    @number_nodes.setter
    def number_nodes(self, value):
        """Set the number of nodes in the mesh"""

        self.generic_data_container.set_property("num_nodes", value)

    @number_elements.setter
    def number_elements(self, value):
        """Set the number of elements in the mesh"""

        self.generic_data_container.set_property("num_elements", value)

    @splittable_by.setter
    def splittable_by(self, value):
        """Set name of the properties according to which the mesh can be split by"""

        self.generic_data_container.set_property("splittable_by", value)

    @available_elem_types.setter
    def available_elem_types(self, value):
        """Set the available element types"""

        self.generic_data_container.set_property("available_elem_types", value)



"""
overlap_fields
==============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class overlap_fields(Operator):
    """Take two fields and superpose them, the overlapping field will
    override values of base_field.

    Parameters
    ----------
    base_field : Field, optional
    overlapping_field : Field, optional


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.overlap_fields()

    >>> # Make input connections
    >>> my_base_field = dpf.Field()
    >>> op.inputs.base_field.connect(my_base_field)
    >>> my_overlapping_field = dpf.Field()
    >>> op.inputs.overlapping_field.connect(my_overlapping_field)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.overlap_fields(
    ...     base_field=my_base_field,
    ...     overlapping_field=my_overlapping_field,
    ... )

    """

    def __init__(
        self, base_field=None, overlapping_field=None, config=None, server=None
    ):
        super().__init__(name="overlap_fields", config=config, server=server)
        self._inputs = InputsOverlapFields(self)
        self._outputs = OutputsOverlapFields(self)
        if base_field is not None:
            self.inputs.base_field.connect(base_field)
        if overlapping_field is not None:
            self.inputs.overlapping_field.connect(overlapping_field)

    @staticmethod
    def _spec():
        description = """Take two fields and superpose them, the overlapping field will
            override values of base_field."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="base_field",
                    type_names=["field"],
                    optional=True,
                    document="""""",
                ),
                1: PinSpecification(
                    name="overlapping_field",
                    type_names=["field"],
                    optional=True,
                    document="""""",
                ),
            },
            map_output_pin_spec={},
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="overlap_fields", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsOverlapFields
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsOverlapFields
        """
        return super().outputs


class InputsOverlapFields(_Inputs):
    """Intermediate class used to connect user inputs to
    overlap_fields operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.overlap_fields()
    >>> my_base_field = dpf.Field()
    >>> op.inputs.base_field.connect(my_base_field)
    >>> my_overlapping_field = dpf.Field()
    >>> op.inputs.overlapping_field.connect(my_overlapping_field)
    """

    def __init__(self, op: Operator):
        super().__init__(overlap_fields._spec().inputs, op)
        self._base_field = Input(overlap_fields._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._base_field)
        self._overlapping_field = Input(overlap_fields._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._overlapping_field)

    @property
    def base_field(self):
        """Allows to connect base_field input to the operator.

        Parameters
        ----------
        my_base_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.overlap_fields()
        >>> op.inputs.base_field.connect(my_base_field)
        >>> # or
        >>> op.inputs.base_field(my_base_field)
        """
        return self._base_field

    @property
    def overlapping_field(self):
        """Allows to connect overlapping_field input to the operator.

        Parameters
        ----------
        my_overlapping_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.overlap_fields()
        >>> op.inputs.overlapping_field.connect(my_overlapping_field)
        >>> # or
        >>> op.inputs.overlapping_field(my_overlapping_field)
        """
        return self._overlapping_field


class OutputsOverlapFields(_Outputs):
    """Intermediate class used to get outputs from
    overlap_fields operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.overlap_fields()
    >>> # Connect inputs : op.inputs. ...
    """

    def __init__(self, op: Operator):
        super().__init__(overlap_fields._spec().outputs, op)



"""
merge_scopings_containers
=========================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class merge_scopings_containers(Operator):
    """Assembles a set of scopings containers into a unique one.

    Parameters
    ----------
    scopings_containers1 : ScopingsContainer
        A vector of scopings containers to merge or
        scopings containers from pin 0 to ...
    scopings_containers2 : ScopingsContainer
        A vector of scopings containers to merge or
        scopings containers from pin 0 to ...


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.merge_scopings_containers()

    >>> # Make input connections
    >>> my_scopings_containers1 = dpf.ScopingsContainer()
    >>> op.inputs.scopings_containers1.connect(my_scopings_containers1)
    >>> my_scopings_containers2 = dpf.ScopingsContainer()
    >>> op.inputs.scopings_containers2.connect(my_scopings_containers2)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.merge_scopings_containers(
    ...     scopings_containers1=my_scopings_containers1,
    ...     scopings_containers2=my_scopings_containers2,
    ... )

    >>> # Get output data
    >>> result_merged_scopings_container = op.outputs.merged_scopings_container()
    """

    def __init__(
        self,
        scopings_containers1=None,
        scopings_containers2=None,
        config=None,
        server=None,
    ):
        super().__init__(name="merge::scopings_container", config=config, server=server)
        self._inputs = InputsMergeScopingsContainers(self)
        self._outputs = OutputsMergeScopingsContainers(self)
        if scopings_containers1 is not None:
            self.inputs.scopings_containers1.connect(scopings_containers1)
        if scopings_containers2 is not None:
            self.inputs.scopings_containers2.connect(scopings_containers2)

    @staticmethod
    def _spec():
        description = """Assembles a set of scopings containers into a unique one."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="scopings_containers",
                    type_names=["scopings_container"],
                    optional=False,
                    document="""A vector of scopings containers to merge or
        scopings containers from pin 0 to ...""",
                ),
                1: PinSpecification(
                    name="scopings_containers",
                    type_names=["scopings_container"],
                    optional=False,
                    document="""A vector of scopings containers to merge or
        scopings containers from pin 0 to ...""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="merged_scopings_container",
                    type_names=["scopings_container"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="merge::scopings_container", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMergeScopingsContainers
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsMergeScopingsContainers
        """
        return super().outputs


class InputsMergeScopingsContainers(_Inputs):
    """Intermediate class used to connect user inputs to
    merge_scopings_containers operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.merge_scopings_containers()
    >>> my_scopings_containers1 = dpf.ScopingsContainer()
    >>> op.inputs.scopings_containers1.connect(my_scopings_containers1)
    >>> my_scopings_containers2 = dpf.ScopingsContainer()
    >>> op.inputs.scopings_containers2.connect(my_scopings_containers2)
    """

    def __init__(self, op: Operator):
        super().__init__(merge_scopings_containers._spec().inputs, op)
        self._scopings_containers1 = Input(
            merge_scopings_containers._spec().input_pin(0), 0, op, 0
        )
        self._inputs.append(self._scopings_containers1)
        self._scopings_containers2 = Input(
            merge_scopings_containers._spec().input_pin(1), 1, op, 1
        )
        self._inputs.append(self._scopings_containers2)

    @property
    def scopings_containers1(self):
        """Allows to connect scopings_containers1 input to the operator.

        A vector of scopings containers to merge or
        scopings containers from pin 0 to ...

        Parameters
        ----------
        my_scopings_containers1 : ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_scopings_containers()
        >>> op.inputs.scopings_containers1.connect(my_scopings_containers1)
        >>> # or
        >>> op.inputs.scopings_containers1(my_scopings_containers1)
        """
        return self._scopings_containers1

    @property
    def scopings_containers2(self):
        """Allows to connect scopings_containers2 input to the operator.

        A vector of scopings containers to merge or
        scopings containers from pin 0 to ...

        Parameters
        ----------
        my_scopings_containers2 : ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_scopings_containers()
        >>> op.inputs.scopings_containers2.connect(my_scopings_containers2)
        >>> # or
        >>> op.inputs.scopings_containers2(my_scopings_containers2)
        """
        return self._scopings_containers2


class OutputsMergeScopingsContainers(_Outputs):
    """Intermediate class used to get outputs from
    merge_scopings_containers operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.merge_scopings_containers()
    >>> # Connect inputs : op.inputs. ...
    >>> result_merged_scopings_container = op.outputs.merged_scopings_container()
    """

    def __init__(self, op: Operator):
        super().__init__(merge_scopings_containers._spec().outputs, op)
        self._merged_scopings_container = Output(
            merge_scopings_containers._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._merged_scopings_container)

    @property
    def merged_scopings_container(self):
        """Allows to get merged_scopings_container output of the operator

        Returns
        ----------
        my_merged_scopings_container : ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.merge_scopings_containers()
        >>> # Connect inputs : op.inputs. ...
        >>> result_merged_scopings_container = op.outputs.merged_scopings_container()
        """  # noqa: E501
        return self._merged_scopings_container



"""Geometry factory module containing functions to create the different geometries."""

import numpy as np

from ansys.dpf.core.geometry import (
    Points,
    Line,
    Plane,
    normalize_vector,
    get_plane_local_axis,
    get_local_coords_from_global,
)


def create_points(coordinates, server=None):
    """Construct points given its coordinates.

    Parameters
    ----------
    coordinates : list, array, Field
        3D coordinates of the points.
    server : :class:`ansys.dpf.core.server`, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    >>> from ansys.dpf.core.geometry_factory import create_points
    >>> points = create_points([[1, 1, 1], [2, 1, 1], [0, 2, 0]])
    >>> print(points)
    DPF Points object:
    Number of points: 3
    Coordinates:
      [1. 1. 1.]
      [2. 1. 1.]
      [0. 2. 0.]
    >>> points.plot()
    """
    return Points(coordinates, server)


def create_line_from_points(points, n_points=100, server=None):
    """Construct line from two points.

    Parameters
    ----------
    points : list, array, Field, Points
        3D coordinates of the points.
    n_points : int
        Number of points in which the line will be discretized.
    server : :class:`ansys.dpf.core.server`, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    >>> from ansys.dpf.core.geometry_factory import create_line_from_points
    >>> line = create_line_from_points([[0, 0, 0], [1, 1, 1]])
    >>> print(line)
    DPF Line object:
    Starting point: [0. 0. 0.]
    Ending point: [1. 1. 1.]
    Line discretized with 100 points
    >>> line.plot()

    """
    return Line(points, n_points, server)


def create_line_from_vector(ini, end=None, n_points=100, server=None):
    """Construct line from origin's coordinates and a vector direction.

    Parameters
    ----------
    ini : list, array, Line
        List 3D coordinates of the initial and ending points of the line.
    end : list, array, Line, optional
        3D coordinates of the ending point of the line (if ``ini`` only contains
        the initial point).
    n_points : int
        Number of points in which the line will be discretized.
    server : :class:`ansys.dpf.core.server`, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    >>> from ansys.dpf.core.geometry_factory import create_line_from_vector
    >>> line = create_line_from_vector([0, 0, 0], [2, 2, 2])
    >>> print(line)
    DPF Line object:
    Starting point: [0. 0. 0.]
    Ending point: [2. 2. 2.]
    Line discretized with 100 points
    >>> line.plot()

    """
    # Input check
    if isinstance(ini[0], list):
        if not len(ini) == 2:
            raise ValueError("Exactly two points must be passed to define the vector")
        if not len(ini[0]) == len(ini[1]) == 3:
            raise ValueError("Each point must contain three coordinates 'x', 'y' and 'z'.")
        vect = ini
    else:
        if not len(ini) == 3:
            raise ValueError(
                "'ini' argument must be of length = 3 \
            representing the 3D coordinates of the origin of the vector."
            )
        if not len(end) == 3:
            raise ValueError(
                "'end' argument must be of length = 3 \
            representing the 3D coordinates of the end point of the vector."
            )
        vect = [ini, end]

    return Line(vect, n_points, server)


def create_plane_from_center_and_normal(
    center, normal, width=1, height=1, n_cells_x=20, n_cells_y=20, server=None
):
    """Create plane from its center and normal direction.

    Parameters
    ----------
    center : list, array, Points
        3D coordinates of the center point of the plane.
    normal : list, array, Line
        Normal direction to the plane.
    width : int, float
        Width of the discretized plane (default = 1).
    height : int, float
        Height of the discretized plane (default = 1).
    n_cells_x : int
        Number of cells in the x direction of the plane (default = 20).
    n_cells_y : int
        Number of cells in the y direction of the plane (default = 20).
    server : :class:`ansys.dpf.core.server`, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    >>> from ansys.dpf.core.geometry_factory import create_plane_from_center_and_normal
    >>> plane = create_plane_from_center_and_normal([1, 1, 1], [0, 0, 1])
    >>> print(plane)
    DPF Plane object:
    Center point: [1, 1, 1]
    Normal direction: [0. 0. 1.]
    Plane discretizaton using:
      Width (x-dir): 1
      Height (y-dir): 1
      Num cells x-dir: 20
      Num cells y-dir: 20
    >>> plane.plot()

    """
    return Plane(center, normal, width, height, n_cells_x, n_cells_y, server)


def create_plane_from_points(points, n_cells_x=20, n_cells_y=20, server=None):
    """Create plane from three points.

    Note that when creating a plane using three points, the plane's width and height
    will be computed such that the three points are the corner points of the plane.

    Parameters
    ----------
    points : list, array, Points
        3D coordinates of the three points defining the plane.
    n_cells_x : int
        Number of cells in the x direction of the plane (default = 20).
    n_cells_y : int
        Number of cells in the y direction of the plane (default = 20).
    server : :class:`ansys.dpf.core.server`, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    >>> from ansys.dpf.core.geometry_factory import create_plane_from_points
    >>> plane = create_plane_from_points([[0, 0, 0], [0, 0, 4], [0, 4, 0]])
    >>> print(plane)
    DPF Plane object:
    Center point: [0.0, 1.3333333333333333, 1.3333333333333333]
    Normal direction: [-1.  0.  0.]
    Plane discretizaton using:
      Width (x-dir): 4.0
      Height (y-dir): 4.0
      Num cells x-dir: 20
      Num cells y-dir: 20
    >>> plane.plot()

    """
    # Input check
    if isinstance(points, Points):
        if not len(points) == 3:
            raise ValueError(
                "Exactly three points must be passed to construct a plane from points."
            )
    else:
        if not len(points) == 3:
            raise ValueError("Exactly three coordinates must be provided to create a plane.")
        if not len(points[0]) == len(points[1]) == len(points[2]) == 3:
            raise ValueError("Each point must contain three coordinates.")

    # Get center and normal from points
    center = get_center_from_coords(points)
    normal_dir = get_normal_direction_from_coords(points)

    # Get width and height from points
    axes_plane = get_plane_local_axis(normal_dir)
    points_local = [get_local_coords_from_global(points[i], axes_plane, center) for i in range(3)]
    x_coords = np.array(points_local)[:, 0]
    y_coords = np.array(points_local)[:, 1]
    width = float(x_coords.max() - x_coords.min())
    height = float(y_coords.max() - y_coords.min())

    return Plane(center, normal_dir, width, height, n_cells_x, n_cells_y, server)


def create_plane_from_lines(
    line1, line2, width=1, height=1, n_cells_x=20, n_cells_y=20, server=None
):
    """Create plane from two lines.

    Parameters
    ----------
    line1 : list, array, Line
        3D coordinates of the two points defining a line.
    line2 : list, array, Line
        3D coordinates of the two points defining a line.
    width : int, float
        Width of the discretized plane (default = 1).
    height : int, float
        Height of the discretized plane (default = 1).
    n_cells_x : int
        Number of cells in the x direction of the plane (default = 20).
    n_cells_y : int
        Number of cells in the y direction of the plane (default = 20).
    server : :class:`ansys.dpf.core.server`, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    >>> from ansys.dpf.core.geometry_factory import create_plane_from_lines
    >>> plane = create_plane_from_lines([[0, 0, 0], [1, 1, 1]], [[0, 1, 0], [2, 1, 0]])
    >>> print(plane)
    DPF Plane object:
    Center point: [-1.5, -0.5, -0.5]
    Normal direction: [ 0.          0.70710678 -0.70710678]
    Plane discretizaton using:
      Width (x-dir): 1
      Height (y-dir): 1
      Num cells x-dir: 20
      Num cells y-dir: 20
    >>> plane.plot()

    """
    # Input check
    if not isinstance(line1, Line) and not isinstance(line2, Line):
        if not len(line1) == len(line2) == 2:
            raise ValueError("Each line must contain two points.")
        if not len(line1[0]) == len(line1[1]) == len(line2[0]) == len(line2[1]) == 3:
            raise ValueError("Each point must contain three coordinates.")

    # Get center and normal from points
    vect1 = [x - y for x, y in zip(line1[0], line1[1])]
    vect2 = [x - y for x, y in zip(line2[0], line2[1])]
    center = get_center_from_coords([vect1, vect2])
    normal = get_cross_product([vect1, vect2])
    return Plane(center, normal, width, height, n_cells_x, n_cells_y, server)


def create_plane_from_point_and_line(
    point, line, width=1, height=1, n_cells_x=20, n_cells_y=20, server=None
):
    """Create plane from point and line.

    Parameters
    ----------
    point : list, array, Points
        3D coordinates of the point.
    line : list, array, Line
        3D coordinates of the two points defining a line.
    width : int, float
        Width of the discretized plane (default = 1).
    height : int, float
        Height of the discretized plane (default = 1).
    n_cells_x : int
        Number of cells in the x direction of the plane (default = 20).
    n_cells_y : int
        Number of cells in the y direction of the plane (default = 20).
    server : :class:`ansys.dpf.core.server`, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    >>> from ansys.dpf.core.geometry_factory import create_plane_from_point_and_line
    >>> plane = create_plane_from_point_and_line([1, 2, 1], [[0, 0, 0], [1, 1, 1]])
    >>> print(plane)
    DPF Plane object:
    Center point: [0.6666666666666666, 1.0, 0.6666666666666666]
    Normal direction: [-0.70710678  0.          0.70710678]
    Plane discretizaton using:
      Width (x-dir): 1
      Height (y-dir): 1
      Num cells x-dir: 20
      Num cells y-dir: 20
    >>> plane.plot()

    """
    # Input check
    if isinstance(point, Points):
        if not len(point) == 1:
            raise ValueError("Only one point must be passed.")
    else:
        if not len(point) == 3:
            raise ValueError("A point must contain three coordinates 'x', 'y' and 'z'.")
    if not isinstance(line, Line):
        if not len(line) == 2:
            raise ValueError("'line' must be of length = 2 containing two points.")
        if not len(line[0]) == len(line[1]) == 3:
            raise ValueError("Each point in line must contain three coordinates 'x', 'y' and 'z'.")

    # Get center and normal from point and vector
    coords = [line[0], line[1], point]
    vects = [line, [line[0], point]]
    center = get_center_from_coords(coords)
    normal = get_cross_product(vects)
    return Plane(center, normal, width, height, n_cells_x, n_cells_y, server)


def get_center_from_coords(coords):
    """Get average coordinates from several points."""
    n_points = len(coords)
    n_coords = len(coords[0])
    return [sum(coords[i][j] for i in range(n_points)) / n_points for j in range(n_coords)]


def get_normal_direction_from_coords(points):
    """Get normal direction between three points."""
    vects = [
        [points[1][i] - points[0][i] for i in range(len(points))],
        [points[2][i] - points[0][i] for i in range(len(points))],
    ]
    return normalize_vector(get_cross_product(vects))


def get_cross_product(vects):
    """Compute cross product between two vectors."""
    return np.cross(vects[0], vects[1])



"""
remote_operator_instantiate
===========================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class remote_operator_instantiate(Operator):
    """Create a local image of an existing remote operator (identified by an
    id and an address) for a given protocol registered in the streams.
    A workflow is created with this operator and returned in output

    Parameters
    ----------
    operator_to_send : int
        Local workflow to push to a remote or id of a
        remote workflow
    output_pin : int
        Pin number of the output to name
    streams_to_remote : StreamsContainer
    data_sources_to_remote : DataSources, optional
    output_name : str
        Output's name of the workflow to return


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.remote_operator_instantiate()

    >>> # Make input connections
    >>> my_operator_to_send = int()
    >>> op.inputs.operator_to_send.connect(my_operator_to_send)
    >>> my_output_pin = int()
    >>> op.inputs.output_pin.connect(my_output_pin)
    >>> my_streams_to_remote = dpf.StreamsContainer()
    >>> op.inputs.streams_to_remote.connect(my_streams_to_remote)
    >>> my_data_sources_to_remote = dpf.DataSources()
    >>> op.inputs.data_sources_to_remote.connect(my_data_sources_to_remote)
    >>> my_output_name = str()
    >>> op.inputs.output_name.connect(my_output_name)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.remote_operator_instantiate(
    ...     operator_to_send=my_operator_to_send,
    ...     output_pin=my_output_pin,
    ...     streams_to_remote=my_streams_to_remote,
    ...     data_sources_to_remote=my_data_sources_to_remote,
    ...     output_name=my_output_name,
    ... )

    >>> # Get output data
    >>> result_remote_workflow = op.outputs.remote_workflow()
    """

    def __init__(
        self,
        operator_to_send=None,
        output_pin=None,
        streams_to_remote=None,
        data_sources_to_remote=None,
        output_name=None,
        config=None,
        server=None,
    ):
        super().__init__(
            name="remote_operator_instantiate", config=config, server=server
        )
        self._inputs = InputsRemoteOperatorInstantiate(self)
        self._outputs = OutputsRemoteOperatorInstantiate(self)
        if operator_to_send is not None:
            self.inputs.operator_to_send.connect(operator_to_send)
        if output_pin is not None:
            self.inputs.output_pin.connect(output_pin)
        if streams_to_remote is not None:
            self.inputs.streams_to_remote.connect(streams_to_remote)
        if data_sources_to_remote is not None:
            self.inputs.data_sources_to_remote.connect(data_sources_to_remote)
        if output_name is not None:
            self.inputs.output_name.connect(output_name)

    @staticmethod
    def _spec():
        description = """Create a local image of an existing remote operator (identified by an
            id and an address) for a given protocol registered in the
            streams. A workflow is created with this operator and
            returned in output"""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="operator_to_send",
                    type_names=["int32"],
                    optional=False,
                    document="""Local workflow to push to a remote or id of a
        remote workflow""",
                ),
                1: PinSpecification(
                    name="output_pin",
                    type_names=["int32"],
                    optional=False,
                    document="""Pin number of the output to name""",
                ),
                3: PinSpecification(
                    name="streams_to_remote",
                    type_names=["streams_container"],
                    optional=False,
                    document="""""",
                ),
                4: PinSpecification(
                    name="data_sources_to_remote",
                    type_names=["data_sources"],
                    optional=True,
                    document="""""",
                ),
                5: PinSpecification(
                    name="output_name",
                    type_names=["string"],
                    optional=False,
                    document="""Output's name of the workflow to return""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="remote_workflow",
                    type_names=["workflow"],
                    optional=False,
                    document="""Remote workflow containing an image of the
        remote workflow and the protocols
        streams""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(
            name="remote_operator_instantiate", server=server
        )

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsRemoteOperatorInstantiate
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsRemoteOperatorInstantiate
        """
        return super().outputs


class InputsRemoteOperatorInstantiate(_Inputs):
    """Intermediate class used to connect user inputs to
    remote_operator_instantiate operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.remote_operator_instantiate()
    >>> my_operator_to_send = int()
    >>> op.inputs.operator_to_send.connect(my_operator_to_send)
    >>> my_output_pin = int()
    >>> op.inputs.output_pin.connect(my_output_pin)
    >>> my_streams_to_remote = dpf.StreamsContainer()
    >>> op.inputs.streams_to_remote.connect(my_streams_to_remote)
    >>> my_data_sources_to_remote = dpf.DataSources()
    >>> op.inputs.data_sources_to_remote.connect(my_data_sources_to_remote)
    >>> my_output_name = str()
    >>> op.inputs.output_name.connect(my_output_name)
    """

    def __init__(self, op: Operator):
        super().__init__(remote_operator_instantiate._spec().inputs, op)
        self._operator_to_send = Input(
            remote_operator_instantiate._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._operator_to_send)
        self._output_pin = Input(
            remote_operator_instantiate._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._output_pin)
        self._streams_to_remote = Input(
            remote_operator_instantiate._spec().input_pin(3), 3, op, -1
        )
        self._inputs.append(self._streams_to_remote)
        self._data_sources_to_remote = Input(
            remote_operator_instantiate._spec().input_pin(4), 4, op, -1
        )
        self._inputs.append(self._data_sources_to_remote)
        self._output_name = Input(
            remote_operator_instantiate._spec().input_pin(5), 5, op, -1
        )
        self._inputs.append(self._output_name)

    @property
    def operator_to_send(self):
        """Allows to connect operator_to_send input to the operator.

        Local workflow to push to a remote or id of a
        remote workflow

        Parameters
        ----------
        my_operator_to_send : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.remote_operator_instantiate()
        >>> op.inputs.operator_to_send.connect(my_operator_to_send)
        >>> # or
        >>> op.inputs.operator_to_send(my_operator_to_send)
        """
        return self._operator_to_send

    @property
    def output_pin(self):
        """Allows to connect output_pin input to the operator.

        Pin number of the output to name

        Parameters
        ----------
        my_output_pin : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.remote_operator_instantiate()
        >>> op.inputs.output_pin.connect(my_output_pin)
        >>> # or
        >>> op.inputs.output_pin(my_output_pin)
        """
        return self._output_pin

    @property
    def streams_to_remote(self):
        """Allows to connect streams_to_remote input to the operator.

        Parameters
        ----------
        my_streams_to_remote : StreamsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.remote_operator_instantiate()
        >>> op.inputs.streams_to_remote.connect(my_streams_to_remote)
        >>> # or
        >>> op.inputs.streams_to_remote(my_streams_to_remote)
        """
        return self._streams_to_remote

    @property
    def data_sources_to_remote(self):
        """Allows to connect data_sources_to_remote input to the operator.

        Parameters
        ----------
        my_data_sources_to_remote : DataSources

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.remote_operator_instantiate()
        >>> op.inputs.data_sources_to_remote.connect(my_data_sources_to_remote)
        >>> # or
        >>> op.inputs.data_sources_to_remote(my_data_sources_to_remote)
        """
        return self._data_sources_to_remote

    @property
    def output_name(self):
        """Allows to connect output_name input to the operator.

        Output's name of the workflow to return

        Parameters
        ----------
        my_output_name : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.remote_operator_instantiate()
        >>> op.inputs.output_name.connect(my_output_name)
        >>> # or
        >>> op.inputs.output_name(my_output_name)
        """
        return self._output_name


class OutputsRemoteOperatorInstantiate(_Outputs):
    """Intermediate class used to get outputs from
    remote_operator_instantiate operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.remote_operator_instantiate()
    >>> # Connect inputs : op.inputs. ...
    >>> result_remote_workflow = op.outputs.remote_workflow()
    """

    def __init__(self, op: Operator):
        super().__init__(remote_operator_instantiate._spec().outputs, op)
        self._remote_workflow = Output(
            remote_operator_instantiate._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._remote_workflow)

    @property
    def remote_workflow(self):
        """Allows to get remote_workflow output of the operator

        Returns
        ----------
        my_remote_workflow : Workflow

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.remote_operator_instantiate()
        >>> # Connect inputs : op.inputs. ...
        >>> result_remote_workflow = op.outputs.remote_workflow()
        """  # noqa: E501
        return self._remote_workflow



"""
skin
====
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class skin(Operator):
    """Extracts a skin of the mesh in a new meshed region. The material ID of
    initial elements are propagated to their facets.

    Parameters
    ----------
    mesh : MeshedRegion
    mesh_scoping : Scoping, optional


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.mesh.skin()

    >>> # Make input connections
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.mesh.skin(
    ...     mesh=my_mesh,
    ...     mesh_scoping=my_mesh_scoping,
    ... )

    >>> # Get output data
    >>> result_mesh = op.outputs.mesh()
    >>> result_nodes_mesh_scoping = op.outputs.nodes_mesh_scoping()
    >>> result_map_new_elements_to_old = op.outputs.map_new_elements_to_old()
    >>> result_property_field_new_elements_to_old = op.outputs.property_field_new_elements_to_old()
    >>> result_facet_indices = op.outputs.facet_indices()
    """

    def __init__(self, mesh=None, mesh_scoping=None, config=None, server=None):
        super().__init__(name="meshed_skin_sector", config=config, server=server)
        self._inputs = InputsSkin(self)
        self._outputs = OutputsSkin(self)
        if mesh is not None:
            self.inputs.mesh.connect(mesh)
        if mesh_scoping is not None:
            self.inputs.mesh_scoping.connect(mesh_scoping)

    @staticmethod
    def _spec():
        description = """Extracts a skin of the mesh in a new meshed region. The material ID of
            initial elements are propagated to their facets."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="mesh_scoping",
                    type_names=["scoping"],
                    optional=True,
                    document="""""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region"],
                    optional=False,
                    document="""Skin meshed region with facets and
        facets_to_ele property fields.""",
                ),
                1: PinSpecification(
                    name="nodes_mesh_scoping",
                    type_names=["scoping"],
                    optional=False,
                    document="""""",
                ),
                2: PinSpecification(
                    name="map_new_elements_to_old",
                    type_names=["umap<int32,int32>"],
                    optional=False,
                    document="""""",
                ),
                3: PinSpecification(
                    name="property_field_new_elements_to_old",
                    type_names=["property_field"],
                    optional=False,
                    document="""This property field provides, for each new
        face element id (in the scoping), the
        corresponding 3d volume element index
        (in the data) it has been extracted
        from. the 3d volume element id can be
        found with the element scoping of the
        input mesh.""",
                ),
                4: PinSpecification(
                    name="facet_indices",
                    type_names=["property_field"],
                    optional=False,
                    document="""This property field gives, for each new face
        element id (in the scoping), the
        corresponding face index on the
        source 3d volume element. the 3d
        volume element can be extracted from
        the previous output.""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="meshed_skin_sector", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsSkin
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsSkin
        """
        return super().outputs


class InputsSkin(_Inputs):
    """Intermediate class used to connect user inputs to
    skin operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mesh.skin()
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    """

    def __init__(self, op: Operator):
        super().__init__(skin._spec().inputs, op)
        self._mesh = Input(skin._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._mesh)
        self._mesh_scoping = Input(skin._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._mesh_scoping)

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator.

        Parameters
        ----------
        my_mesh : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.skin()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> # or
        >>> op.inputs.mesh(my_mesh)
        """
        return self._mesh

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator.

        Parameters
        ----------
        my_mesh_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.skin()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> # or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)
        """
        return self._mesh_scoping


class OutputsSkin(_Outputs):
    """Intermediate class used to get outputs from
    skin operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mesh.skin()
    >>> # Connect inputs : op.inputs. ...
    >>> result_mesh = op.outputs.mesh()
    >>> result_nodes_mesh_scoping = op.outputs.nodes_mesh_scoping()
    >>> result_map_new_elements_to_old = op.outputs.map_new_elements_to_old()
    >>> result_property_field_new_elements_to_old = op.outputs.property_field_new_elements_to_old()
    >>> result_facet_indices = op.outputs.facet_indices()
    """

    def __init__(self, op: Operator):
        super().__init__(skin._spec().outputs, op)
        self._mesh = Output(skin._spec().output_pin(0), 0, op)
        self._outputs.append(self._mesh)
        self._nodes_mesh_scoping = Output(skin._spec().output_pin(1), 1, op)
        self._outputs.append(self._nodes_mesh_scoping)
        self._map_new_elements_to_old = Output(skin._spec().output_pin(2), 2, op)
        self._outputs.append(self._map_new_elements_to_old)
        self._property_field_new_elements_to_old = Output(
            skin._spec().output_pin(3), 3, op
        )
        self._outputs.append(self._property_field_new_elements_to_old)
        self._facet_indices = Output(skin._spec().output_pin(4), 4, op)
        self._outputs.append(self._facet_indices)

    @property
    def mesh(self):
        """Allows to get mesh output of the operator

        Returns
        ----------
        my_mesh : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.skin()
        >>> # Connect inputs : op.inputs. ...
        >>> result_mesh = op.outputs.mesh()
        """  # noqa: E501
        return self._mesh

    @property
    def nodes_mesh_scoping(self):
        """Allows to get nodes_mesh_scoping output of the operator

        Returns
        ----------
        my_nodes_mesh_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.skin()
        >>> # Connect inputs : op.inputs. ...
        >>> result_nodes_mesh_scoping = op.outputs.nodes_mesh_scoping()
        """  # noqa: E501
        return self._nodes_mesh_scoping

    @property
    def map_new_elements_to_old(self):
        """Allows to get map_new_elements_to_old output of the operator

        Returns
        ----------
        my_map_new_elements_to_old :

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.skin()
        >>> # Connect inputs : op.inputs. ...
        >>> result_map_new_elements_to_old = op.outputs.map_new_elements_to_old()
        """  # noqa: E501
        return self._map_new_elements_to_old

    @property
    def property_field_new_elements_to_old(self):
        """Allows to get property_field_new_elements_to_old output of the operator

        Returns
        ----------
        my_property_field_new_elements_to_old : PropertyField

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.skin()
        >>> # Connect inputs : op.inputs. ...
        >>> result_property_field_new_elements_to_old = op.outputs.property_field_new_elements_to_old()
        """  # noqa: E501
        return self._property_field_new_elements_to_old

    @property
    def facet_indices(self):
        """Allows to get facet_indices output of the operator

        Returns
        ----------
        my_facet_indices : PropertyField

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.skin()
        >>> # Connect inputs : op.inputs. ...
        >>> result_facet_indices = op.outputs.facet_indices()
        """  # noqa: E501
        return self._facet_indices



"""
Dimensionality
==============
"""
from ansys.dpf.core.common import natures


class Dimensionality:
    """Represents the dimensionality of a field.

        This class reads a list of dimension vectors and creates a dimensionality message.

    Parameters
    ----------
    dim_vec : list
        List of integers for the dimension vectors. There is a 1D vector for a
        scalar or 3D vector. There is a 2D vector for a matrix. For example:

        * ``[1]``: scalar
        * ``[3]``: 3d vector
        * ``[3, 3]``: ``3 x 3`` matrix
    nature : :class:`ansys.dpf.core.common.natures`

    """

    def __init__(self, dim_vec=None, nature: natures = natures.vector):
        self.dim = dim_vec
        self.nature = nature
        # set nature
        if not isinstance(nature, natures):
            if isinstance(nature, str):
                self.nature = natures[nature]
            elif isinstance(nature, int):
                self.nature = natures(nature)
            else:
                raise TypeError("Nature is expected to be cast into ansys.dpf.core.natures enum")

        if self.dim is not None and 0 in self.dim:
            self.dim = [x for x in self.dim if x != 0]
            if len(self.dim) == 0:
                self.dim = None

        if self.dim is None:
            if self.nature == natures.vector:
                self.dim = [3]
            elif self.nature == natures.symmatrix:
                self.dim = [3, 3]
            elif self.nature == natures.scalar:
                self.dim = [1]

    def is_1d_dim(self):
        return len(self.dim) == 1

    def is_2d_dim(self):
        return len(self.dim) == 2

    def __str__(self):
        return str(self.dim) + " " + self.nature.name

    @property
    def component_count(self):
        count = 1
        for comp in self.dim:
            count *= comp
        return count

    @staticmethod
    def scalar_dim():
        """Retrieve the dimensionality of a scalar field.

        Returns
        -------
        :class:`ansys.dpf.core.dimensionnality.Dimensionality`
            Dimensionality of the scalar field.
        """
        return Dimensionality([1], natures.scalar)

    @staticmethod
    def vector_dim(size):
        """Retrieve the dimensionality of a vector field of a given size.

        Parameters
        ----------
        size : int
            Number of components by entity.

        Returns
        -------
        :class:`ansys.dpf.core.dimensionnality.Dimensionality`
            Dimensionality of the vector field.
        """
        return Dimensionality([size], natures.vector)

    @staticmethod
    def vector_3d_dim():
        """Retrieve the dimensionality of a three-dimension vector field.

        Returns
        -------
        :class:`ansys.dpf.core.dimensionnality.Dimensionality`
            Dimensionality of the three-dimension vector field.
        """
        return Dimensionality([3], natures.vector)

    @staticmethod
    def tensor_dim():
        """Retrieve the dimensionality of a symmetrical ``3 x 3`` tensor field.

        Returns
        -------
        :class:`ansys.dpf.core.dimensionnality.Dimensionality`
            Dimensionality of the symmetrical ``3 x 3`` tensor field.
        """
        return Dimensionality([3, 3], natures.symmatrix)

    def __eq__(self, other):
        return self.dim == other.dim and self.nature == other.nature

    def __ne__(self, other):
        return not self.__eq__(other)



"""
.. _ref_support:

Support
=======
"""
import traceback
import warnings
from ansys.dpf.gate import support_capi, support_grpcapi
from ansys.dpf.core.check_version import version_requires

from ansys.dpf.core import server as server_module
from ansys.dpf.core import collection


class Support:
    """Base class for support (supporting Field's location, Scoping's location,
    Collection's labels...)

    Field, PropertyField and StringField support can be accessed generically via this base class.

    Parameters
    ----------
    support : ctypes.c_void_p, ansys.grpc.dpf.support_pb2.Support
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Examples
    --------
    Create a time frequency support from a model.

    >>> from ansys.dpf.core import Model
    >>> from ansys.dpf.core import examples
    >>> transient = examples.download_transient_result()
    >>> model = Model(transient)
    >>> time_freq_support = model.metadata.time_freq_support # printable
    >>> time_freq_support.available_field_supported_properties()
    ['time_freqs']
    >>> field = time_freq_support.field_support_by_property("time_freqs")

    """

    def __init__(self, support, server=None):
        # step 1: get server
        self._server = server_module.get_or_create_server(server)

        if not self._server.meet_version("3.0"):  # pragma: no cover
            return

        # step 2: get api
        self._support_api = self._server.get_api_for_type(
            capi=support_capi.SupportCAPI, grpcapi=support_grpcapi.SupportGRPCAPI
        )

        # step3: init environment
        self._support_api.init_support_environment(self)  # creates stub when gRPC

        # step4: take instance:
        # object_name -> protobuf.message, DPFObject*
        self._internal_obj = support

    @version_requires("5.0")
    def field_support_by_property(self, property_name: str):
        """Returns a Field supporting (describing) a given property.

        Returns
        -------
        Field

        Notes
        -----
        Available with server's version starting at 5.0.
        """
        from ansys.dpf.core.field import Field

        out = self._support_api.support_get_field_support_by_property(self, property_name)
        if out is not None:
            return Field(field=out, server=self._server)

    @version_requires("5.0")
    def prop_field_support_by_property(self, property_name: str):
        """Returns a PropertyField supporting (describing) a given property.

        Returns
        -------
        PropertyField

        Notes
        -----
        Available with server's version starting at 5.0.
        """
        from ansys.dpf.core.property_field import PropertyField

        out = self._support_api.support_get_property_field_support_by_property(self, property_name)
        if out is not None:
            return PropertyField(property_field=out, server=self._server)

    @version_requires("5.0")
    def string_field_support_by_property(self, property_name: str):
        """Returns a StringField supporting (describing) a given property.

        Returns
        -------
        StringField

        Notes
        -----
        Available with server's version starting at 5.0.
        """
        from ansys.dpf.core.string_field import StringField

        out = self._support_api.support_get_string_field_support_by_property(self, property_name)
        if out is not None:
            return StringField(string_field=out, server=self._server)

    @version_requires("5.0")
    def available_field_supported_properties(self):
        """Returns the list of property names supported by a Field.

        Returns
        -------
        list

        Notes
        -----
        Available with server's version starting at 5.0.
        """
        coll_obj = collection.StringCollection(
            collection=self._support_api.support_get_property_names_as_string_coll_for_fields(self),
            server=self._server,
        )
        return coll_obj.get_integral_entries()

    @version_requires("5.0")
    def available_prop_field_supported_properties(self):
        """Returns the list of property names supported by a PropertyField.

        Returns
        -------
        list

        Notes
        -----
        Available with server's version starting at 5.0.
        """
        coll_obj = collection.StringCollection(
            collection=self._support_api.support_get_property_names_as_string_coll_for_property_fields(  # noqa: E501
                self
            ),
            server=self._server,
        )
        return coll_obj.get_integral_entries()

    @version_requires("5.0")
    def available_string_field_supported_properties(self):
        """Returns the list of property names supported by a StringField.

        Returns
        -------
        list

        Notes
        -----
        Available with server's version starting at 5.0.
        """
        coll_obj = collection.StringCollection(
            collection=self._support_api.support_get_property_names_as_string_coll_for_string_fields(  # noqa: E501
                self
            ),
            server=self._server,
        )
        return coll_obj.get_integral_entries()

    def __del__(self):
        try:
            self._deleter_func[0](self._deleter_func[1](self))
        except:
            warnings.warn(traceback.format_exc())



"""
StringField
===========
"""

import numpy as np
from ansys.dpf.core.common import natures, locations, _get_size_of_list
from ansys.dpf.core import scoping
from ansys.dpf.core import server as server_module
from ansys.dpf.core import errors
from ansys.dpf.core.field_base import _FieldBase
from ansys.dpf.gate import (
    string_field_abstract_api,
    string_field_capi,
    string_field_grpcapi,
    dpf_vector,
    integral_types,
)


class StringField(_FieldBase):
    """Describes string data scoped on entities such as names.

    This class is a field with string values instead of double values.

    Parameters
    ----------
    nentities: int
        Number of entities that the string field is to contain (reserved). The
        default is ``0``.
    string_field : Field, ansys.grpc.dpf.field_pb2.Field, ctypes.c_void_p, optional
        Field message generated from a gRPC stub, or returned by DPF's C clients.
    server : server.DPFServer, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> pfield = dpf.StringField()
    >>> list_ids = [1, 2, 4, 6, 7]
    >>> scop = dpf.Scoping(ids = list_ids, location = dpf.locations.nodal)
    >>> pfield.scoping = scop
    >>> list_data = ["water", "oil", "gaz", "paint", "air"]
    >>> pfield.data = list_data

    Notes
    -----
    Class available with server's version starting at 5.0 (Ansys 2023R1).
    """

    def __init__(
        self,
        nentities=0,
        string_field=None,
        server=None,
    ):
        self._server = server_module.get_or_create_server(server)
        if string_field is None and not self._server.meet_version("5.0"):
            raise errors.DpfVersionNotSupported("5.0")
        super().__init__(
            nentities=nentities,
            nature=natures.scalar,
            field=string_field,
            server=server,
        )

    @property
    def _api(self) -> string_field_abstract_api.StringFieldAbstractAPI:
        if not self._api_instance:
            self._api_instance = self._server.get_api_for_type(
                capi=string_field_capi.StringFieldCAPI,
                grpcapi=string_field_grpcapi.StringFieldGRPCAPI,
            )
        return self._api_instance

    def _init_api_env(self):
        self._api.init_string_field_environment(self)

    @staticmethod
    def _field_create_internal_obj(
        api: string_field_abstract_api.StringFieldAbstractAPI,
        client,
        nature,
        nentities,
        location=locations.nodal,
        ncomp_n=0,
        ncomp_m=0,
        with_type=None,
    ):
        if client is not None:
            return api.csstring_field_new_on_client(client, nentities, nentities)
        else:
            return api.csstring_field_new(nentities, nentities)

    @property
    def location(self):
        """Location of the string field.

        A property field contains a scoping, which is the location that is read.
        To update location, directly update the scoping location.

        Returns
        -------
        location : str
            Location string, can be found in ``dpf.locations``: ie.
            ``dpf.locations.nodal`` or ``dpf.locations.elemental``.

        Examples
        --------
        Create a string field and request the location.

        >>> from ansys.dpf import core as dpf
        >>> pfield = dpf.StringField()
        >>> list_ids = [1, 2, 4, 6, 7]
        >>> scop = dpf.Scoping(ids = list_ids, location = dpf.locations.nodal)
        >>> pfield.scoping = scop
        >>> pfield.scoping.location = dpf.locations.nodal
        >>> pfield.location
        'Nodal'

        """
        if self.scoping:
            return self.scoping.location
        else:
            return None

    @location.setter
    def location(self, value):
        """Change the property field location.

        Parameters
        ----------
        location : str or locations
            Location string in :class:`ansys.dpf.core.locations`.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> pfield = dpf.StringField()
        >>> scop = dpf.Scoping(ids = list_ids, location = dpf.locations.nodal)
        >>> pfield.scoping = scop
        >>> pfield.location = 'Nodal'
        >>> pfield.location
        'Nodal'

        """
        if self.scoping:
            self.scoping.location = value
        else:
            raise Exception(
                "Property field location is based on scoping, and scoping is not defined"
            )

    @property
    def component_count(self):
        return 1

    @property
    def elementary_data_count(self):
        return self._api.csstring_field_get_data_size(self)

    @property
    def size(self):
        return self._api.csstring_field_get_data_size(self)

    def _set_scoping(self, scoping):
        self._api.csstring_field_set_cscoping(self, scoping)

    def _get_scoping(self):
        return scoping.Scoping(
            scoping=self._api.csstring_field_get_cscoping(self), server=self._server
        )

    def get_entity_data(self, index):
        try:
            vec = dpf_vector.DPFVectorString(client=self._server.client)
            self._api.csstring_field_get_entity_data_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size, index
            )
            return vec
        except NotImplementedError:
            data = self._api.csstring_field_get_entity_data(self, index)
            return data

    def get_entity_data_by_id(self, id):
        try:
            vec = dpf_vector.DPFVectorString(client=self._server.client)
            self._api.csstring_field_get_entity_data_by_id_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size, id
            )
            return vec
        except NotImplementedError:
            index = self.scoping.index(id)
            if index < 0:
                raise ValueError(f"The ID {id} must be greater than 0.")
            data = self.get_entity_data(index)
            return data

    def append(self, data, scopingid):
        string_list = integral_types.MutableListString(data)
        self._api.csstring_field_push_back(self, scopingid, _get_size_of_list(data), string_list)

    def _get_data(self, np_array=True):
        try:
            vec = dpf_vector.DPFVectorString(client=self._server.client)
            self._api.csstring_field_get_data_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size
            )
            return vec
        except NotImplementedError:
            data = self._api.csstring_field_get_data(self, np_array)
            return data

    def _set_data(self, data):
        if isinstance(data, (np.ndarray, np.generic)):
            if (
                0 != self.size
                and self.component_count > 1
                and data.size // self.component_count != data.size / self.component_count
            ):
                raise ValueError(
                    f"An array of shape {self.shape} is expected and "
                    f"shape {data.shape} was input"
                )
        string_list = integral_types.MutableListString(data)
        return self._api.csstring_field_set_data(self, _get_size_of_list(data), string_list)



"""
.. _ref_incremental:

Incremental
===========
"""

from ansys.dpf import core

from typing import Dict, Any


class IncrementalHelper:
    """Provides an API to transform an existing workflow into an incrementally evaluating one.

    It works by plugging operators into an incomplete workflow.

    Example
    -------
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> path = examples.find_msup_transient()
    >>> ds = dpf.DataSources(path)
    >>> scoping = dpf.time_freq_scoping_factory.scoping_on_all_time_freqs(ds)
    >>>
    >>> result_op = dpf.operators.result.displacement(data_sources=ds, time_scoping=scoping)
    >>> minmax_op = dpf.operators.min_max.min_max_fc_inc(result_op)
    >>>
    >>> new_op = dpf.split_workflow_in_chunks(result_op, minmax_op, scoping, chunk_size=5)
    >>> min_field = new_op.get_output(0, dpf.types.field)
    >>> max_field = new_op.get_output(1, dpf.types.field)
    """

    def __init__(
        self,
        start_op: core.Operator,
        end_op: core.Operator,
        scoping: core.Scoping,
        scoping_pin: int = None,
    ):
        """Constructs an IncrementalHelper object.

        Given the first and the last operator of a workflow, as well as the scoping.

        This class can be used to simplify the use of incremental operators, and automatically
        be enabled to incrementally evaluate a workflow.

        Under the constraint that the end_op supports incremental evaluation.

        Parameters
        ----------
        start_op : Operator
            First operator in the workflow to convert
        end_op : Operator
            Last operator in the workflow to convert (Operator providing the meaningful output)
        scoping : Scoping
            Scoping used to chunk the data
        scoping_pin : int, optional
            Pin number of the scoping on the first operator, otherwise it is deduced with types
        """
        # Input operator should accept a scoping
        # Last operator should support incremental evaluation
        # but as we don't have a consistent method to check,
        # it should be permissive in the case the specification isn't up to date
        self._start_op = start_op
        self._end_op = self._map_to_incremental(end_op)

        self._scoping = scoping
        self._scoping_pin = self._find_scoping_pin(scoping_pin)

    def estimate_size(self, max_bytes: int, _dict_inputs: Dict[int, Any] = {}) -> int:
        """Estimates the chunk size from the estimated number of bytes outputted in one iteration.

        Estimation is based on the size of the output for one ID of the given time_scoping,
        so it will run the operator for only one iteration.

        It only supports Field and FieldContainer.
        For other types, you should specify chunk_size argument in the split() method.

        Parameters
        ----------
            max_bytes : int
                Max allowed size of an output from the first operator, for one iteration (in bytes).
            _dict_inputs: dict[int,any]
                Dictionary associating pin number to inputs, for evaluating output of one iteration.
        """
        # Evaluate for the first element to try to guess memory consumption
        # It is best to use with a lot of elements
        first_id = self._scoping.ids[0]
        srv = self._scoping._server
        loc = self._scoping.location
        _dict_inputs[self._scoping_pin] = core.Scoping(server=srv, ids=[first_id], location=loc)

        outputs = self._prerun(_dict_inputs)

        _outputs = outputs._outputs
        data = map(lambda o: o.get_data(), _outputs)
        # Output sizes of all inputs for one iteration
        sizes = map(lambda obj: self._compute_size(obj), data)

        # Total size for one ID in the scoping
        size_for_one = sum(sizes)
        # total_size = size_for_one * self._scoping.size

        num_iter = int(max_bytes / size_for_one)
        num_iter = min(max(num_iter, 1), self._scoping.size)  # clamp(num_iter, 1, scoping size)
        return num_iter  # math.gcd(num_iter, self._scoping.size)

    def _compute_size(self, obj):
        if isinstance(obj, core.FieldsContainer):
            fc = obj
            return self._compute_size(fc[0])
        elif isinstance(obj, core.Field):
            field = obj
            # Double = 8 bytes assumption
            return field.size * 8

        raise NotImplementedError()

    def _prerun(self, _dict_inputs: Dict[int, Any]):
        """"""

        for pin_idx, val in _dict_inputs.items():
            self._start_op.connect(pin_idx, val)
        self._start_op.run()
        return self._start_op.outputs

    # Transforms a user workflow:
    #
    #           +----------+    +---------------+    +---------+
    # scoping ->| start_op | -> | middle ops... | -> | end_op  | ->
    #           +----------+    +---------------+    +---------+
    #
    # Into a new workflow like this:
    #
    #           +----------+    +---------------+    +---------+
    # scoping ->| start_op | -> | middle ops... | -> | end_op  |
    #   \       +----------+    +---------------+    +---------+
    #    \           \                                  |
    #     \           \   +------------------+          |                    (pins remaps)
    #      \           \> |                  |          |   +----------+    +-----------+
    #       \ scop_pin -> | chunk_in         |          +-> |          | -> | forward   | -> final
    #        +----------> |   for_each_range |  iterables   |          |    | (new end) |    outputs
    #       chunk_size -> |                  | -----------> | for_each |    +-----------+
    #                     +------------------+              |          |
    #                                     end_input_pin-->  |          |
    #                                                       +----------+
    def split(
        self, chunk_size: int, end_input_pin: int = 0, rescope: bool = False
    ) -> core.Operator:
        """Integrate given operators into a new workflow enabling incremental evaluation.

        Given a chunk size (multiple of given scoping), it will provide a new operator to retrieve
        outputs from, and enable incremental evaluation, notably reducing peak memory usage.

        Parameters
        ----------
            chunk_size : int
                Number of iterations per run
            end_input_pin : int, optional
                Pin number of the output to use from the first operator (default = 0)
            rescope : bool, optional
                Rescope all the outputs based on the given scoping (default = False)
        """
        # Enables incremental evaluation:
        # Using for_each, chunk_in_for_each_range and incremental version of the last operator
        # by returning two operators with remapped inputs and outputs to other operators

        _server = self._start_op._server

        for_each = core.Operator("for_each", server=_server)
        split_in_range = core.Operator("chunk_in_for_each_range", server=_server)
        forward = core.Operator("forward", server=_server)

        split_in_range.connect_operator_as_input(1, self._start_op)
        split_in_range.connect(2, self._scoping_pin)
        split_in_range.connect(3, self._scoping)
        split_in_range.connect(4, chunk_size)

        for_each.connect(0, split_in_range, 0)
        for_each.connect(2, end_input_pin)

        # connect inputs
        dict_outputs = core.Operator.operator_specification(
            op_name=self._end_op.name, server=_server
        ).outputs
        if not dict_outputs:
            # temporary patch for incremental:: operators
            dict_outputs = {0: None}

        fe_pin_idx = 3  # see doc of for_each
        for pin_idx in dict_outputs.keys():
            # connect end_op to for_each
            for_each.connect(fe_pin_idx, self._end_op, pin_idx)
            # remap
            forward.connect(pin_idx, for_each, fe_pin_idx)
            fe_pin_idx += 1

        output = forward

        if rescope:
            new_forward = core.Operator("forward")
            for pin_idx in dict_outputs.keys():
                rescope = core.Operator("Rescope")
                rescope.connect(0, forward, pin_idx)
                rescope.connect(1, self._scoping)
                new_forward.connect(pin_idx, rescope, 0)

            output = new_forward
        return output

    def _map_to_incremental(self, end_op: core.Operator):
        # The goal of this function is to validate that a given operator is indeed incremental.
        # If an operator is found to not support incremental evaluation, it must not be strict
        # it should only output warnings
        # because this function -by design- may be outdated.
        inc_operators = [
            "accumulate_level_over_label_fc",
            "accumulate_min_over_label_fc",
            "accumulate_over_label_fc",
            "average_over_label_fc",
            "min_max_inc",
            "min_max_fc_inc",
            "max_over_time_by_entity",
            "min_max_over_time_by_entity",
            "min_max_by_time",
            "min_over_time_by_entity",
            "time_of_max_by_entity",
            "time_of_min_by_entity",
            "incremental::merge::property_field",
            "incremental::merge::mesh",
            "incremental::merge::field",
            "incremental::merge::fields_container",
        ]

        map_to_inc = {"min_max": "min_max_inc", "min_max_fc": "min_max_fc_inc"}

        if end_op.name not in inc_operators:
            print(f"WARNING: Operator named {end_op.name} may not support incremental evaluation")
            if end_op.name in map_to_inc.keys():
                print(
                    f"An operator named {map_to_inc[end_op.name]} supports incremental evaluation"
                )

        if "incremental" in end_op.config.available_config_options:
            end_op.config.set_config_option("incremental", True)

        return end_op

    def _find_scoping_pin(self, pin_idx):
        dict_inputs = self._start_op.inputs._dict_inputs
        # validate given pin_idx
        if pin_idx != None and pin_idx in dict_inputs:
            pin_spec = dict_inputs[pin_idx]
            if "scoping" in pin_spec.type_names:
                return pin_idx

        # look for scoping pin
        for pin_idx, spec in dict_inputs.items():
            if "scoping" in spec.type_names:
                return pin_idx

        raise Exception(
            f"Scoping pin could not be found in start_op with name '{self._start_op.name}'"
        )


def split_workflow_in_chunks(
    start_op: core.Operator,
    end_op: core.Operator,
    scoping: core.Scoping,
    rescope: bool = False,
    max_bytes: int = 1024**3,
    dict_inputs: Dict[int, Any] = {},
    chunk_size: int = None,
    scoping_pin: int = None,
    end_input_pin: int = 0,
):
    """Transforms a workflow into an incrementally evaluating one.

    It wraps in one method the functionality of the IncrementalHelper class as well
    as the estimation of the chunk size.

    If no chunk_size is specified, the function will attempt to estimate the value
    by calling IncrementalHelper.estimate_size(max_bytes, dict_inputs).

    If no scoping_pin is specified, the function will attempt to deduce the correct pin,
    which would be the first input pin matching a scoping type.

    Parameters
    ----------
        start_op : Operator
            Initial operator of the workflow to convert
        end_op : Operator
            Last operator of the workflow to convert
        scoping : Scoping
            Scoping to split across multiple evaluation
        rescope : bool, optional
            If enabled, will rescope final outputs with the given scoping (default = False)
        max_bytes : int, optional
            Max allowed size for the output from the first operator (default = 1024**3)
        dict_inputs : dict[int, any], optional
            Inputs to pass to the first operator, used only for the estimation run (default = {})
        chunk_size = int, optional
            Maximum number of scoping elements to process in an iteration (default = None)
        scoping_pin : int, optional
            The pin number on the first operator to bind the scoping (default = None)
        end_input_pin : int, optional
            Pin number of the output to use from the first operator(default = 0)
    """
    splitter = IncrementalHelper(start_op, end_op, scoping, scoping_pin)

    if chunk_size == None:
        print(f"Estimating chunk_size with max_bytes: {max_bytes}")
        chunk_size = splitter.estimate_size(max_bytes, dict_inputs)
        print(f"Done. chunk_size set to {chunk_size} (scoping size: {scoping.size})")

    return splitter.split(chunk_size, end_input_pin, rescope)



"""
euler_load_buckling
===================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class euler_load_buckling(Operator):
    """Computing Euler's Critical Load. Formula: Ncr = n*E*I*pi*pi /(L*L)

    Parameters
    ----------
    field_beam_end_condition : DataSources or Field
        This pin contains file csv or field of beam's
        end condition added by the user. if
        there's no file added, it would take
        value of all beam's end condition as
        1.
    field_beam_moment_inertia : Field
        Field of beam's moment inertia
    field_beam_young_modulus : Field
        Field of beam's young modulus
    field_beam_length : Field
        Field of beam's length


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.result.euler_load_buckling()

    >>> # Make input connections
    >>> my_field_beam_end_condition = dpf.DataSources()
    >>> op.inputs.field_beam_end_condition.connect(my_field_beam_end_condition)
    >>> my_field_beam_moment_inertia = dpf.Field()
    >>> op.inputs.field_beam_moment_inertia.connect(my_field_beam_moment_inertia)
    >>> my_field_beam_young_modulus = dpf.Field()
    >>> op.inputs.field_beam_young_modulus.connect(my_field_beam_young_modulus)
    >>> my_field_beam_length = dpf.Field()
    >>> op.inputs.field_beam_length.connect(my_field_beam_length)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.result.euler_load_buckling(
    ...     field_beam_end_condition=my_field_beam_end_condition,
    ...     field_beam_moment_inertia=my_field_beam_moment_inertia,
    ...     field_beam_young_modulus=my_field_beam_young_modulus,
    ...     field_beam_length=my_field_beam_length,
    ... )

    >>> # Get output data
    >>> result_field_euler_critical_load = op.outputs.field_euler_critical_load()
    >>> result_field_euler_critical_load_yy = op.outputs.field_euler_critical_load_yy()
    >>> result_field_euler_critical_load_zz = op.outputs.field_euler_critical_load_zz()
    """

    def __init__(
        self,
        field_beam_end_condition=None,
        field_beam_moment_inertia=None,
        field_beam_young_modulus=None,
        field_beam_length=None,
        config=None,
        server=None,
    ):
        super().__init__(name="euler_load_buckling", config=config, server=server)
        self._inputs = InputsEulerLoadBuckling(self)
        self._outputs = OutputsEulerLoadBuckling(self)
        if field_beam_end_condition is not None:
            self.inputs.field_beam_end_condition.connect(field_beam_end_condition)
        if field_beam_moment_inertia is not None:
            self.inputs.field_beam_moment_inertia.connect(field_beam_moment_inertia)
        if field_beam_young_modulus is not None:
            self.inputs.field_beam_young_modulus.connect(field_beam_young_modulus)
        if field_beam_length is not None:
            self.inputs.field_beam_length.connect(field_beam_length)

    @staticmethod
    def _spec():
        description = (
            """Computing Euler's Critical Load. Formula: Ncr = n*E*I*pi*pi /(L*L)"""
        )
        spec = Specification(
            description=description,
            map_input_pin_spec={
                5: PinSpecification(
                    name="field_beam_end_condition",
                    type_names=["data_sources", "field"],
                    optional=False,
                    document="""This pin contains file csv or field of beam's
        end condition added by the user. if
        there's no file added, it would take
        value of all beam's end condition as
        1.""",
                ),
                6: PinSpecification(
                    name="field_beam_moment_inertia",
                    type_names=["field"],
                    optional=False,
                    document="""Field of beam's moment inertia""",
                ),
                7: PinSpecification(
                    name="field_beam_young_modulus",
                    type_names=["field"],
                    optional=False,
                    document="""Field of beam's young modulus""",
                ),
                8: PinSpecification(
                    name="field_beam_length",
                    type_names=["field"],
                    optional=False,
                    document="""Field of beam's length""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field_euler_critical_load",
                    type_names=["field"],
                    optional=False,
                    document="""This field contains euler's critical load
        about the principle axis of the cross
        section having the least moment of
        inertia.""",
                ),
                1: PinSpecification(
                    name="field_euler_critical_load_yy",
                    type_names=["field"],
                    optional=False,
                    document="""This field contains euler's critical load on
        axis y.""",
                ),
                2: PinSpecification(
                    name="field_euler_critical_load_zz",
                    type_names=["field"],
                    optional=False,
                    document="""This field contains euler's critical load on
        axis z.""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="euler_load_buckling", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsEulerLoadBuckling
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsEulerLoadBuckling
        """
        return super().outputs


class InputsEulerLoadBuckling(_Inputs):
    """Intermediate class used to connect user inputs to
    euler_load_buckling operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.euler_load_buckling()
    >>> my_field_beam_end_condition = dpf.DataSources()
    >>> op.inputs.field_beam_end_condition.connect(my_field_beam_end_condition)
    >>> my_field_beam_moment_inertia = dpf.Field()
    >>> op.inputs.field_beam_moment_inertia.connect(my_field_beam_moment_inertia)
    >>> my_field_beam_young_modulus = dpf.Field()
    >>> op.inputs.field_beam_young_modulus.connect(my_field_beam_young_modulus)
    >>> my_field_beam_length = dpf.Field()
    >>> op.inputs.field_beam_length.connect(my_field_beam_length)
    """

    def __init__(self, op: Operator):
        super().__init__(euler_load_buckling._spec().inputs, op)
        self._field_beam_end_condition = Input(
            euler_load_buckling._spec().input_pin(5), 5, op, -1
        )
        self._inputs.append(self._field_beam_end_condition)
        self._field_beam_moment_inertia = Input(
            euler_load_buckling._spec().input_pin(6), 6, op, -1
        )
        self._inputs.append(self._field_beam_moment_inertia)
        self._field_beam_young_modulus = Input(
            euler_load_buckling._spec().input_pin(7), 7, op, -1
        )
        self._inputs.append(self._field_beam_young_modulus)
        self._field_beam_length = Input(
            euler_load_buckling._spec().input_pin(8), 8, op, -1
        )
        self._inputs.append(self._field_beam_length)

    @property
    def field_beam_end_condition(self):
        """Allows to connect field_beam_end_condition input to the operator.

        This pin contains file csv or field of beam's
        end condition added by the user. if
        there's no file added, it would take
        value of all beam's end condition as
        1.

        Parameters
        ----------
        my_field_beam_end_condition : DataSources or Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.euler_load_buckling()
        >>> op.inputs.field_beam_end_condition.connect(my_field_beam_end_condition)
        >>> # or
        >>> op.inputs.field_beam_end_condition(my_field_beam_end_condition)
        """
        return self._field_beam_end_condition

    @property
    def field_beam_moment_inertia(self):
        """Allows to connect field_beam_moment_inertia input to the operator.

        Field of beam's moment inertia

        Parameters
        ----------
        my_field_beam_moment_inertia : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.euler_load_buckling()
        >>> op.inputs.field_beam_moment_inertia.connect(my_field_beam_moment_inertia)
        >>> # or
        >>> op.inputs.field_beam_moment_inertia(my_field_beam_moment_inertia)
        """
        return self._field_beam_moment_inertia

    @property
    def field_beam_young_modulus(self):
        """Allows to connect field_beam_young_modulus input to the operator.

        Field of beam's young modulus

        Parameters
        ----------
        my_field_beam_young_modulus : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.euler_load_buckling()
        >>> op.inputs.field_beam_young_modulus.connect(my_field_beam_young_modulus)
        >>> # or
        >>> op.inputs.field_beam_young_modulus(my_field_beam_young_modulus)
        """
        return self._field_beam_young_modulus

    @property
    def field_beam_length(self):
        """Allows to connect field_beam_length input to the operator.

        Field of beam's length

        Parameters
        ----------
        my_field_beam_length : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.euler_load_buckling()
        >>> op.inputs.field_beam_length.connect(my_field_beam_length)
        >>> # or
        >>> op.inputs.field_beam_length(my_field_beam_length)
        """
        return self._field_beam_length


class OutputsEulerLoadBuckling(_Outputs):
    """Intermediate class used to get outputs from
    euler_load_buckling operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.euler_load_buckling()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field_euler_critical_load = op.outputs.field_euler_critical_load()
    >>> result_field_euler_critical_load_yy = op.outputs.field_euler_critical_load_yy()
    >>> result_field_euler_critical_load_zz = op.outputs.field_euler_critical_load_zz()
    """

    def __init__(self, op: Operator):
        super().__init__(euler_load_buckling._spec().outputs, op)
        self._field_euler_critical_load = Output(
            euler_load_buckling._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._field_euler_critical_load)
        self._field_euler_critical_load_yy = Output(
            euler_load_buckling._spec().output_pin(1), 1, op
        )
        self._outputs.append(self._field_euler_critical_load_yy)
        self._field_euler_critical_load_zz = Output(
            euler_load_buckling._spec().output_pin(2), 2, op
        )
        self._outputs.append(self._field_euler_critical_load_zz)

    @property
    def field_euler_critical_load(self):
        """Allows to get field_euler_critical_load output of the operator

        Returns
        ----------
        my_field_euler_critical_load : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.euler_load_buckling()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field_euler_critical_load = op.outputs.field_euler_critical_load()
        """  # noqa: E501
        return self._field_euler_critical_load

    @property
    def field_euler_critical_load_yy(self):
        """Allows to get field_euler_critical_load_yy output of the operator

        Returns
        ----------
        my_field_euler_critical_load_yy : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.euler_load_buckling()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field_euler_critical_load_yy = op.outputs.field_euler_critical_load_yy()
        """  # noqa: E501
        return self._field_euler_critical_load_yy

    @property
    def field_euler_critical_load_zz(self):
        """Allows to get field_euler_critical_load_zz output of the operator

        Returns
        ----------
        my_field_euler_critical_load_zz : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.euler_load_buckling()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field_euler_critical_load_zz = op.outputs.field_euler_critical_load_zz()
        """  # noqa: E501
        return self._field_euler_critical_load_zz



"""
.. _ref_custom_type_field:

CustomTypeField
===============
"""
import warnings

import numpy as np

from ansys.dpf.core import server as server_module
from ansys.dpf.core import errors
from ansys.dpf.core import scoping
from ansys.dpf.core.common import locations, _get_size_of_list
from ansys.dpf.core.field_base import _FieldBase
from ansys.dpf.core.field_definition import FieldDefinition
from ansys.dpf.core.support import Support
from ansys.dpf.gate import (
    dpf_array,
    dpf_vector,
    integral_types,
)


class dict_with_missing_numpy_type(dict):
    def __missing__(self, key):
        return key.name


numpy_type_to_dpf = dict_with_missing_numpy_type(
    {
        np.float64: "double",
        np.float32: "float",
        np.int16: "short",
        np.byte: "char",
        np.int8: "char",
    }
)


class CustomTypeField(_FieldBase):
    """Represents a simulation data container with each unitary data being of a custom type.
    When initializing the ``CustomTypeField`` class, provide a unitary data type.
    The ``CustomTypeField`` class gives you the ability to choose the most optimized unitary
    data type for a given usage, and hence, allows you to optimize memory usage.

    This can be evaluated data from the :class:`Operator <ansys.dpf.core.Operator>` class
    or created directly by an instance of this class.

    A field's data is always associated to its scoping (entities
    associated to each value) and support (subset of the model where the
    data is), making the field a self-describing piece of data.

    Parameters
    ----------
    unitary_type : numpy.dtype
        The data vector of the Field will be a vector of this custom unitary type.
    nentities : int, optional
        Number of entities reserved. The default is ``0``.
    field : CustomTypeField, ansys.grpc.dpf.field_pb2.Field, ctypes.c_void_p, optional
        Field message generated from a gRPC stub, or returned by DPF's C clients.
    server : :class:`ansys.dpf.core.server`, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    Create a custom type field from scratch.

    >>> from ansys.dpf.core import locations
    >>> from ansys.dpf import core as dpf
    >>> import numpy as np
    >>> field = dpf.CustomTypeField(unitary_type=np.uint64)
    >>> field.location = locations.nodal
    >>> field.append([1000000,2000000], 1)
    >>> float_field = dpf.CustomTypeField(unitary_type=np.float32)
    >>> float_field.is_of_type(np.float32)
    True

    Notes
    -----
    Class available with server's version starting at 5.0 (Ansys 2023R1).
    """

    def __init__(
        self,
        unitary_type=None,
        nentities=0,
        field=None,
        server=None,
    ):
        """Initialize the field either with an optional field message or
        by connecting to a stub.
        """
        self._server = server_module.get_or_create_server(server)
        if field is None and not self._server.meet_version("5.0"):
            raise errors.DpfVersionNotSupported("5.0")
        if unitary_type is not None:
            self._type = np.dtype(unitary_type)
        else:
            self._type = unitary_type
        super().__init__(nentities=nentities, field=field, server=server)
        self._load_type()
        self._field_definition = self._load_field_definition()

    def _load_type(self):
        if self._type is None:
            type_name = integral_types.MutableString(256)
            unitary_size = integral_types.MutableInt32(0)
            self._api.cscustom_type_field_get_type(self, type_name, unitary_size)
            try:
                self._type = np.dtype(str(type_name))
                if self._type.itemsize != int(unitary_size):
                    warnings.warn(
                        f"The field unitary data type will be interpreted as a void, of "
                        f"{int(unitary_size)} bytes because"
                        f" a {str(type_name)} type was expected to have "
                        f"{self._type.itemsize} bytes."
                    )
                    raise TypeError()
            except TypeError:
                self._type = np.dtype(f"V{int(unitary_size)}")

    @property
    def _api(self):
        from ansys.dpf.gate import custom_type_field_capi, custom_type_field_grpcapi

        if not self._api_instance:
            self._api_instance = self._server.get_api_for_type(
                capi=custom_type_field_capi.CustomTypeFieldCAPI,
                grpcapi=custom_type_field_grpcapi.CustomTypeFieldGRPCAPI,
            )
        return self._api_instance

    def _init_api_env(self):
        self._api.init_custom_type_field_environment(self)

    @staticmethod
    def _field_create_internal_obj(
        api,
        client,
        nature,
        nentities,
        location=locations.nodal,
        ncomp_n=0,
        ncomp_m=0,
        with_type=None,
    ):
        dpf_type_name = numpy_type_to_dpf[with_type]
        if client is not None:
            return api.cscustom_type_field_new_on_client(
                client, dpf_type_name, with_type.itemsize, nentities, nentities
            )
        else:
            return api.cscustom_type_field_new(
                dpf_type_name, with_type.itemsize, nentities, nentities
            )

    @property
    def location(self):
        """CustomTypeField location.

        Returns
        -------
        str
            Location string, which can be ``"Nodal"``, ``"Elemental"``,
            ``"ElementalNodal"``... See :class:`ansys.dpf.core.common.locations`.

        Examples
        --------
        >>> from ansys.dpf.core import locations
        >>> from ansys.dpf import core as dpf
        >>> import numpy as np
        >>> field = dpf.CustomTypeField(unitary_type=np.uint64)
        >>> field.location = locations.nodal
        >>> field.location
        'Nodal'

        """
        if self.field_definition:
            return self.field_definition.location

    @location.setter
    def location(self, location):
        """Change the field location.

        Parameters
        -------
        location : str or locations
             Location string, which can be ``"Nodal"``, ``"Elemental"``,
            ``"ElementalNodal"``... See :class:`ansys.dpf.core.common.locations`.

        Examples
        --------
        >>> from ansys.dpf.core import locations
        >>> from ansys.dpf import core as dpf
        >>> import numpy as np
        >>> field = dpf.CustomTypeField(unitary_type=np.uint64)
        >>> field.location = locations.nodal
        >>> field.location
        'Nodal'

        """
        fielddef = self.field_definition
        fielddef.location = location
        self.field_definition = fielddef

    def is_of_type(self, type_to_compare: np.dtype) -> bool:
        """Checks whether the Field's unitary type is the same as the input type.

        Parameters
        ----------
        type_to_compare: numpy.dtype

        Returns
        -------
        bool

        Examples
        --------
        Create a custom type field from scratch.

        >>> from ansys.dpf import core as dpf
        >>> import numpy as np
        >>> field = dpf.CustomTypeField(unitary_type=np.int16)
        >>> field.is_of_type(np.int16)
        True
        >>> field.is_of_type(np.short)
        True
        >>> field.is_of_type(np.int32)
        False

        """
        return self.type == type_to_compare

    @property
    def type(self):
        """Type of unitary data in the Field's data vector.
        Should be properly set at the Field construction to have properly allocated data.

        Returns
        -------
        numpy.dtype

        Examples
        --------
        >>> from ansys.dpf.core import locations
        >>> from ansys.dpf import core as dpf
        >>> import numpy as np
        >>> field = dpf.CustomTypeField(unitary_type=np.uint64)
        >>> field.type
        dtype('uint64')

        """
        return self._type

    @property
    def component_count(self):
        return self._api.cscustom_type_field_get_number_of_components(self)

    @property
    def elementary_data_count(self):
        return self._api.cscustom_type_field_get_number_elementary_data(self)

    @property
    def size(self):
        return self._api.cscustom_type_field_get_data_size(self)

    def _set_scoping(self, scoping):
        self._api.cscustom_type_field_set_cscoping(self, scoping)

    def _get_scoping(self):
        obj = self._api.cscustom_type_field_get_cscoping(self)
        if obj is not None:
            return scoping.Scoping(scoping=obj, server=self._server)

    def get_entity_data(self, index):
        """Returns the array corresponding to the data of a given entity index.

        Parameters
        ----------
        index: int
            Index in the ``Scoping``.

        Returns
        -------
        numpy.ndarray

        Examples
        --------
        >>> from ansys.dpf.core import locations
        >>> from ansys.dpf import core as dpf
        >>> import numpy as np
        >>> field = dpf.CustomTypeField(unitary_type=np.uint64)
        >>> field.append([1000000, 2000000], 1)
        >>> field.append([1000000, 2000000, 3000000], 2)
        >>> field.get_entity_data(0)
        DPFArray([1000000, 2000000]...
        >>> field.get_entity_data(1)
        DPFArray([1000000, 2000000, 3000000]...

        """
        try:
            vec = dpf_vector.DPFVectorCustomType(self._type, client=self._server.client)
            self._api.cscustom_type_field_get_entity_data_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size, index
            )
            data = dpf_array.DPFArray(vec)

        except NotImplementedError:
            data = self._api.cscustom_type_field_get_entity_data(self, index)
        n_comp = self.component_count
        if n_comp != 1 and data.size != 0:
            data.shape = (data.size // n_comp, n_comp)
        return data

    def get_entity_data_by_id(self, id):
        """Returns the array corresponding to the data of a given entity id.

        Parameters
        ----------
        id: int
            Entity ID in the ``Scoping``.

        Returns
        -------
        numpy.ndarray

        Examples
        --------
        >>> from ansys.dpf.core import locations
        >>> from ansys.dpf import core as dpf
        >>> import numpy as np
        >>> field = dpf.CustomTypeField(unitary_type=np.uint64)
        >>> field.append([1000000, 2000000], 1)
        >>> field.append([1000000, 2000000, 3000000], 2)
        >>> field.get_entity_data_by_id(1)
        DPFArray([1000000, 2000000]...
        >>> field.get_entity_data_by_id(2)
        DPFArray([1000000, 2000000, 3000000]...

        """
        try:
            vec = dpf_vector.DPFVectorCustomType(self._type, client=self._server.client)
            self._api.cscustom_type_field_get_entity_data_by_id_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size, id
            )
            data = dpf_array.DPFArray(vec)

        except NotImplementedError:
            index = self.scoping.index(id)
            if index < 0:
                raise ValueError(f"The ID {id} must be greater than 0.")
            data = self.get_entity_data(index)
        n_comp = self.component_count
        if n_comp != 1 and data.size != 0:
            data.shape = (data.size // n_comp, n_comp)
        return data

    def append(self, data, scopingid):
        if isinstance(data, list):
            data = np.array(data, dtype=self._type)
        self._api.cscustom_type_field_push_back(self, scopingid, _get_size_of_list(data), data)

    def _get_data_pointer(self):
        try:
            vec = dpf_vector.DPFVectorInt(client=self._server.client)
            self._api.cscustom_type_field_get_data_pointer_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size
            )
            return dpf_array.DPFArray(vec)

        except NotImplementedError:
            return self._api.cscustom_type_field_get_data_pointer(self, True)

    def _set_data_pointer(self, data):
        return self._api.cscustom_type_field_set_data_pointer(self, _get_size_of_list(data), data)

    def _get_data(self, np_array=True):
        try:
            vec = dpf_vector.DPFVectorCustomType(self._type, client=self._server.client)
            self._api.cscustom_type_field_get_data_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size
            )
            data = dpf_array.DPFArray(vec) if np_array else dpf_array.DPFArray(vec).tolist()
        except NotImplementedError:
            data = self._api.cscustom_type_field_get_data(self, np_array)
        n_comp = self.component_count
        if np_array and n_comp != 1 and data.size != 0:
            data.shape = (data.size // n_comp, n_comp)
        return data

    def _set_data(self, data):
        if isinstance(data, (np.ndarray, np.generic)):
            if data.dtype != self._type:
                data = data.astype(self._type)
        else:
            data = np.array(data, dtype=self._type)
        size = _get_size_of_list(data)
        return self._api.cscustom_type_field_set_data(self, size, data)

    def resize(self, nentities, datasize):
        """Allocate memory.

        Parameters
        ----------
        nentities : int
            Number of IDs in the scoping.
        datasize : int
            Size of the data vector.

        """
        return self._api.cscustom_type_field_resize(self, datasize, nentities)

    def _load_field_definition(self):
        """Attempt to load the field definition for this field."""
        # try:
        out = self._api.cscustom_type_field_get_shared_field_definition(self)
        return FieldDefinition(out, self._server)

    @property
    def unit(self):
        """Units for the field.

        Returns
        ----------
        str
           Units for the field.

        Examples
        --------
        Units for a displacement field.

        >>> from ansys.dpf import core as dpf
        >>> my_field = dpf.CustomTypeField(int, 10)
        >>> my_field.unit = "m"
        >>> my_field.unit
        'm'

        """
        if self.field_definition:
            return self.field_definition.unit

    @unit.setter
    def unit(self, value):
        """Change the unit for the field

        Parameters
        ----------
        value : str
            Units for the field.

        Examples
        --------
        Units for a displacement field.

        >>> from ansys.dpf import core as dpf
        >>> my_field = dpf.CustomTypeField(10, dpf.natures.vector,dpf.locations.nodal)
        >>> my_field.unit = "m"
        >>> my_field.unit
        'm'

        """
        fielddef = self.field_definition
        fielddef.unit = value
        self.field_definition = fielddef

    @property
    def dimensionality(self):
        """Dimensionality represents the shape of the elementary data contained in the field.

        Returns
        -------
        dimensionality : :class:`ansys.dpf.core.dimensionality.Dimensionality`
            Nature and size of the elementary data.
        """
        if self.field_definition:
            return self.field_definition.dimensionality

    @dimensionality.setter
    def dimensionality(self, value):
        fielddef = self.field_definition
        fielddef.dimensionality = value
        self.field_definition = fielddef

    @property
    def name(self):
        """Name of the field."""
        return self._field_definition.name

    def _set_field_definition(self, field_definition):
        """Set the field definition.

        Parameters
        ----------
        field_definition : :class"`ansys.dpf.core.field_definition.FieldDefinition`

        """
        self._api.cscustom_type_field_set_field_definition(self, field_definition)
        self._field_definition = self._load_field_definition()

    @property
    def field_definition(self):
        """CustomTypeField information, including its location, unit, dimensionality
        and shell layers.

        Returns
        -------
        :class:`ansys.dpf.core.field_definition.FieldDefinition`

        """
        return self._field_definition

    @field_definition.setter
    def field_definition(self, value):
        self._set_field_definition(value)

    def _set_support(self, support):
        self._api.cscustom_type_field_set_support(self, support)

    @property
    def support(self):
        obj = self._api.cscustom_type_field_get_support(self)
        if obj is not None:
            return Support(support=obj, server=self._server)

    @support.setter
    def support(self, val):
        self._set_support(val)



# -*- coding: utf-8 -*-
"""
StreamsContainer
================
Contains classes associated with the DPF StreamsContainer.
"""
import warnings
import traceback

from ansys.dpf.core.server_types import BaseServer
from ansys.dpf.core import server as server_module
from ansys.dpf.gate import (
    streams_capi,
    data_processing_capi,
    data_processing_grpcapi,
)
from ansys.dpf.core import errors


class StreamsContainer:
    """Python wrapper for operator input or output of streams.

    Streams define open, ready-to-use, data sources.
    Once the files in the streams are opened, they stay opened and
    they keep some data in cache to make the next evaluations faster.
    To close the opened files, release the handles.

    Note
    ----
    Only available for an InProcess server configuration.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.find_multishells_rst())
    >>> streams_provider = model.metadata.streams_provider
    >>> sc = streams_provider.outputs.streams_container()
    """

    def __init__(self, streams_container=None, server: BaseServer = None):
        # step 1: get server
        self._server = server_module.get_or_create_server(server)
        if self._server.has_client():
            txt = """
            StreamsContainer only available for server with InProcess communication protocol
            """
            raise errors.ServerTypeError(txt)

        # step2: if object exists, take the instance, else create it
        self._internal_obj = None
        if streams_container is not None:
            if isinstance(streams_container, StreamsContainer):
                self._server = streams_container._server
                core_api = self._server.get_api_for_type(
                    capi=data_processing_capi.DataProcessingCAPI,
                    grpcapi=data_processing_grpcapi.DataProcessingGRPCAPI,
                )
                core_api.init_data_processing_environment(self)
                self._internal_obj = core_api.data_processing_duplicate_object_reference(
                    streams_container
                )
            else:
                self._internal_obj = streams_container
        self.owned = False

    @property
    def _server(self):
        return self._server_instance

    @_server.setter
    def _server(self, value):
        self._server_instance = value
        # step 2: get api
        self._api = self._server.get_api_for_type(capi=streams_capi.StreamsCAPI, grpcapi=None)
        if not self._api:
            raise NotImplementedError("StreamsContainer unavailable for LegacyGrpc servers")
        # step3: init environment
        self._api.init_streams_environment(self)  # creates stub when gRPC

    def release_handles(self):
        """Release the streams."""
        self._api.streams_release_handles(self)

    def __del__(self):
        """Delete the entry."""
        try:
            # delete
            if not self.owned:
                self._deleter_func[0](self._deleter_func[1](self))
        except:  # pylint: disable=bare-except
            warnings.warn(traceback.format_exc())



"""
.. _ref_data_sources:

Data Sources
============
"""
import os
import warnings
import traceback

from ansys.dpf.core import server as server_module
from ansys.dpf.gate import (
    data_sources_capi,
    data_sources_grpcapi,
    integral_types,
    data_processing_capi,
    data_processing_grpcapi,
)

from ansys.dpf.core.check_version import version_requires


class DataSources:
    """Contains files with analysis results.

    An extension key (``'rst'`` for example) is used to choose which files represent
    results files versus accessory files. You can set a result file path when
    initializing this class.


    Parameters
    ----------
    result_path : str or os.PathLike object, optional
        Path of the result. The default is ``None``.
    data_sources : ansys.grpc.dpf.data_sources_pb2.DataSources
        gRPC data sources message. The default is ``None``.
    server : server.DPFServer, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    Initialize a model from a result path.

    >>> from ansys.dpf import core as dpf
    >>> my_data_sources = dpf.DataSources('file.rst')
    >>> my_data_sources.result_files
    ['file.rst']

    """

    def __init__(self, result_path=None, data_sources=None, server=None):
        """Initialize a connection with the server."""
        # step 1: get server
        self._server = server_module.get_or_create_server(server)

        # step 2: get api
        self._api = self._server.get_api_for_type(
            capi=data_sources_capi.DataSourcesCAPI,
            grpcapi=data_sources_grpcapi.DataSourcesGRPCAPI,
        )

        # step3: init environment
        self._api.init_data_sources_environment(self)  # creates stub when gRPC

        # step4: if object exists: take instance, else create it:
        # object_name -> protobuf.message, DPFObject*
        if data_sources is not None:
            if isinstance(data_sources, DataSources):
                # Make a Copy
                core_api = self._server.get_api_for_type(
                    capi=data_processing_capi.DataProcessingCAPI,
                    grpcapi=data_processing_grpcapi.DataProcessingGRPCAPI,
                )
                core_api.init_data_processing_environment(self)
                self._internal_obj = core_api.data_processing_duplicate_object_reference(
                    data_sources
                )
            else:
                # It should be a message (usually from a call to operator_getoutput_data_sources)
                self._internal_obj = data_sources
        else:
            if self._server.has_client():
                self._internal_obj = self._api.data_sources_new_on_client(self._server.client)
            else:
                self._internal_obj = self._api.data_sources_new("data_sources")

        if result_path is not None:
            self.set_result_file_path(result_path)

    def set_result_file_path(self, filepath, key=""):
        """Add a result file path to the data sources.

        Parameters
        ----------
        filepath : str or os.PathLike object
            Path to the result file.
        key : str, optional
            Extension of the file, which is used as a key for choosing the correct
            plugin when a result is requested by an operator. The default is ``""``,
            in which case the key is found directly.

        Examples
        --------
        Create a data source and set the result file path.

        >>> from ansys.dpf import core as dpf
        >>> data_sources = dpf.DataSources()
        >>> data_sources.set_result_file_path('/tmp/file.rst')
        >>> data_sources.result_files
        ['/tmp/file.rst']

        """
        if key == "":
            self._api.data_sources_set_result_file_path_utf8(self, str(filepath))
        else:
            self._api.data_sources_set_result_file_path_with_key_utf8(self, str(filepath), key)

    def set_domain_result_file_path(self, path, domain_id):
        """Add a result file path by domain.

        This method is used to handle files created by a
        distributed solve.

        Parameters
        ----------
        path: str or os.PathLike object
            Path to the file.
        domain_id: int
            Domain ID for the distributed files.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_sources = dpf.DataSources()
        >>> data_sources.set_domain_result_file_path('/tmp/file0.sub', 0)
        >>> data_sources.set_domain_result_file_path('/tmp/file1.sub', 1)

        """
        self._api.data_sources_set_domain_result_file_path_utf8(self, str(path), domain_id)

    def add_file_path(self, filepath, key="", is_domain: bool = False, domain_id=0):
        """Add a file path to the data sources.

        Files not added as result files are accessory files, which contain accessory
        information not present in the result files.

        Parameters
        ----------
        filepath : str or os.PathLike object
            Path of the file.
        key : str, optional
            Extension of the file, which is used as a key for choosing the correct
            plugin when a result is requested by an operator. The default is ``""``,
            in which case the key is found directly.
        is_domain: bool, optional
            Whether the file path is the domain path. The default is ``False``.
        domain_id: int, optional
            Domain ID for the distributed files. The default is ``0``. For this
            parameter to be taken into account, ``domain_path=True`` must be set.
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_sources = dpf.DataSources()
        >>> data_sources.add_file_path('/tmp/ds.dat')

        """
        # The filename needs to be a fully qualified file name
        if not os.path.dirname(filepath):
            # append local path
            filepath = os.path.join(os.getcwd(), os.path.basename(filepath))
        if is_domain:
            if key == "":
                raise NotImplementedError("A key must be given when using is_domain=True.")
            else:
                self._api.data_sources_add_domain_file_path_with_key_utf8(
                    self, str(filepath), key, domain_id
                )
        else:
            if key == "":
                self._api.data_sources_add_file_path_utf8(self, str(filepath))
            else:
                self._api.data_sources_add_file_path_with_key_utf8(self, str(filepath), key)

    def add_file_path_for_specified_result(self, filepath, key="", result_key=""):
        """Add a file path for a specified result file key to the data sources.

        This method can be used when results files with different keys (extensions) are
        contained in the data sources. For example, a solve using two different solvers
        could generate two different sets of files.

        Parameters
        ----------
        filepath : str or os.PathLike object
            Path of the file.
        key : str, optional
            Extension of the file, which is used as a key for choosing the correct
            plugin when a result is requested by an operator. The default is ``""``,
            in which case the key is found directly.
        result_key: str, optional
            Extension of the results file that the specified file path belongs to.
            The default is ``""``, in which case the key is found directly.
        """
        # The filename needs to be a fully qualified file name
        if not os.path.dirname(filepath):
            # append local path
            filepath = os.path.join(os.getcwd(), os.path.basename(filepath))

        self._api.data_sources_add_file_path_for_specified_result_utf8(
            self, str(filepath), key, result_key
        )

    def add_upstream(self, upstream_data_sources, result_key=""):
        """Add upstream data sources.

        This is used to add a set of path creating an upstream for
        recursive workflows.

        Parameters
        ----------
        upstream_data_sources : DataSources
            Set of paths creating an upstream for recursive workflows.

        result_key: str, optional
            Extension of the result file group with which this upstream belongs

        """
        if result_key == "":
            self._api.data_sources_add_upstream_data_sources(self, upstream_data_sources)
        else:
            self._api.data_sources_add_upstream_data_sources_for_specified_result(
                self, upstream_data_sources, result_key
            )

    def add_upstream_for_domain(self, upstream_data_sources, domain_id):
        """Add an upstream data sources for a given domain.

        This is used to add a set of path creating an upstream for
        recursive workflows in a distributed solve.

        Parameters
        ----------
        upstream_data_sources : DataSources
            Set of paths creating an upstream for recursive workflows.

        domain_id: int
            Domain id for distributed files.

        """
        self._api.data_sources_add_upstream_domain_data_sources(
            self, upstream_data_sources, domain_id
        )

    @property
    def result_key(self):
        """Result key used by the data sources.

        Returns
        -------
        str
           Result key.

        """
        return self._api.data_sources_get_result_key(self)

    @property
    def result_files(self):
        """List of result files contained in the data sources.

        Returns
        ----------
        list
            List of result files.
        """
        result_key = self.result_key
        if result_key == "":
            return None
        else:
            response = []
            num_keys = self._api.data_sources_get_num_keys(self)
            for i_key in range(num_keys):
                num_paths = integral_types.MutableInt32()
                key = self._api.data_sources_get_key(self, i_key, num_paths)
                if key == result_key:
                    for i_path in range(int(num_paths)):
                        path = self._api.data_sources_get_path(self, key, i_path)
                        response.append(path)
            return response

    @version_requires("7.0")
    def register_namespace(self, result_key: str, namespace: str):
        """Adds a link from this ``result_key`` to this ``namespace`` in the DataSources.
        This ``result_key`` to ``namespace`` mapping is used by source operators
        to find internal operators to call.

        Notes
        -----
        Available with server's version starting at 7.0.
        """
        self._api.data_sources_register_namespace(self, result_key, namespace)

    def __str__(self):
        """Describe the entity.

        Returns
        -------
        str
            Description of the entity.
        """
        from ansys.dpf.core.core import _description

        return _description(self._internal_obj, self._server)

    def __del__(self):
        try:
            self._deleter_func[0](self._deleter_func[1](self))
        except:
            warnings.warn(traceback.format_exc())
            pass



"""
mesh_plan_clip
==============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class mesh_plan_clip(Operator):
    """Clips a volume mesh along a plane and keeps one side.

    Parameters
    ----------
    mesh_or_field : MeshedRegion or Field
    normal : Field
        An overall 3d vector that gives the normal
        direction of the plane.
    origin : Field
        An overall 3d vector that gives a point of
        the plane.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.mesh.mesh_plan_clip()

    >>> # Make input connections
    >>> my_mesh_or_field = dpf.MeshedRegion()
    >>> op.inputs.mesh_or_field.connect(my_mesh_or_field)
    >>> my_normal = dpf.Field()
    >>> op.inputs.normal.connect(my_normal)
    >>> my_origin = dpf.Field()
    >>> op.inputs.origin.connect(my_origin)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.mesh.mesh_plan_clip(
    ...     mesh_or_field=my_mesh_or_field,
    ...     normal=my_normal,
    ...     origin=my_origin,
    ... )

    >>> # Get output data
    >>> result_field = op.outputs.field()
    >>> result_mesh = op.outputs.mesh()
    """

    def __init__(
        self, mesh_or_field=None, normal=None, origin=None, config=None, server=None
    ):
        super().__init__(name="mesh_plan_clip", config=config, server=server)
        self._inputs = InputsMeshPlanClip(self)
        self._outputs = OutputsMeshPlanClip(self)
        if mesh_or_field is not None:
            self.inputs.mesh_or_field.connect(mesh_or_field)
        if normal is not None:
            self.inputs.normal.connect(normal)
        if origin is not None:
            self.inputs.origin.connect(origin)

    @staticmethod
    def _spec():
        description = """Clips a volume mesh along a plane and keeps one side."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="mesh_or_field",
                    type_names=["abstract_meshed_region", "field"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="normal",
                    type_names=["field"],
                    optional=False,
                    document="""An overall 3d vector that gives the normal
        direction of the plane.""",
                ),
                2: PinSpecification(
                    name="origin",
                    type_names=["field"],
                    optional=False,
                    document="""An overall 3d vector that gives a point of
        the plane.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
                2: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="mesh_plan_clip", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMeshPlanClip
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsMeshPlanClip
        """
        return super().outputs


class InputsMeshPlanClip(_Inputs):
    """Intermediate class used to connect user inputs to
    mesh_plan_clip operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mesh.mesh_plan_clip()
    >>> my_mesh_or_field = dpf.MeshedRegion()
    >>> op.inputs.mesh_or_field.connect(my_mesh_or_field)
    >>> my_normal = dpf.Field()
    >>> op.inputs.normal.connect(my_normal)
    >>> my_origin = dpf.Field()
    >>> op.inputs.origin.connect(my_origin)
    """

    def __init__(self, op: Operator):
        super().__init__(mesh_plan_clip._spec().inputs, op)
        self._mesh_or_field = Input(mesh_plan_clip._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._mesh_or_field)
        self._normal = Input(mesh_plan_clip._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._normal)
        self._origin = Input(mesh_plan_clip._spec().input_pin(2), 2, op, -1)
        self._inputs.append(self._origin)

    @property
    def mesh_or_field(self):
        """Allows to connect mesh_or_field input to the operator.

        Parameters
        ----------
        my_mesh_or_field : MeshedRegion or Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.mesh_plan_clip()
        >>> op.inputs.mesh_or_field.connect(my_mesh_or_field)
        >>> # or
        >>> op.inputs.mesh_or_field(my_mesh_or_field)
        """
        return self._mesh_or_field

    @property
    def normal(self):
        """Allows to connect normal input to the operator.

        An overall 3d vector that gives the normal
        direction of the plane.

        Parameters
        ----------
        my_normal : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.mesh_plan_clip()
        >>> op.inputs.normal.connect(my_normal)
        >>> # or
        >>> op.inputs.normal(my_normal)
        """
        return self._normal

    @property
    def origin(self):
        """Allows to connect origin input to the operator.

        An overall 3d vector that gives a point of
        the plane.

        Parameters
        ----------
        my_origin : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.mesh_plan_clip()
        >>> op.inputs.origin.connect(my_origin)
        >>> # or
        >>> op.inputs.origin(my_origin)
        """
        return self._origin


class OutputsMeshPlanClip(_Outputs):
    """Intermediate class used to get outputs from
    mesh_plan_clip operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mesh.mesh_plan_clip()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field = op.outputs.field()
    >>> result_mesh = op.outputs.mesh()
    """

    def __init__(self, op: Operator):
        super().__init__(mesh_plan_clip._spec().outputs, op)
        self._field = Output(mesh_plan_clip._spec().output_pin(0), 0, op)
        self._outputs.append(self._field)
        self._mesh = Output(mesh_plan_clip._spec().output_pin(2), 2, op)
        self._outputs.append(self._mesh)

    @property
    def field(self):
        """Allows to get field output of the operator

        Returns
        ----------
        my_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.mesh_plan_clip()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field()
        """  # noqa: E501
        return self._field

    @property
    def mesh(self):
        """Allows to get mesh output of the operator

        Returns
        ----------
        my_mesh : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.mesh_plan_clip()
        >>> # Connect inputs : op.inputs. ...
        >>> result_mesh = op.outputs.mesh()
        """  # noqa: E501
        return self._mesh



"""
fields_factory
==============

Contains functions to simplify creating fields.
"""

from ansys.dpf.core.common import natures, locations
from ansys.dpf.core import Field
from ansys.dpf.core import server as server_module
from ansys.dpf.gate import field_capi, field_grpcapi

import numpy as np


def field_from_array(arr, server=None):
    """Create a DPF vector or scalar field from a numpy array or a Python list.

    Parameters
    ----------
    arr : np.ndarray or List
        Numpy array or Python list containing either 1 or 3 dimensions.

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        Field constructed from the array.
    """
    from ansys.dpf.core import Field, natures

    arr = np.asarray(arr)

    if not np.issubdtype(arr.dtype, np.number):
        raise TypeError("Array must be a numeric type")

    shp_err = ValueError(
        "Array must be either contain 1 dimension or " "2 dimensions with three components."
    )
    if arr.ndim == 1:
        nature = natures.scalar
    elif arr.ndim == 2:
        if arr.shape[1] == 1:
            arr = arr.ravel()
            nature = natures.scalar
        elif arr.shape[1] == 3:
            nature = natures.vector
        elif arr.shape[1] == 6:
            nature = natures.symmatrix
        else:
            raise shp_err
    else:
        raise shp_err

    n_entities = arr.shape[0]
    field = Field(nentities=n_entities, nature=nature, server=server)
    field.data = arr
    field.scoping.ids = np.arange(1, n_entities + 1)
    return field


def create_matrix_field(num_entities, num_lines, num_col, location=locations.nodal, server=None):
    """Create a matrix :class:`ansys.dpf.core.Field`.

    This field contains entities that have a matrix format. This is a "reserve" mechanism,
    not a resize one. This means that you need to append data to grow the size of your field.

    Parameters
    ----------
    num_entities : int
        Number of entities to reserve.
    num_lines : int
        Number of matrix line.
    num_col : int
        Number of matrix columns.
    location : str, optional
        Location of the field. Options are in :class:`locations <ansys.dpf.core.common.locations>`.
        The default is ``dpf.locations.nodal``.

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field of the requested format.

    Examples
    --------
    Create a field containing 3 matrix entities of a col*lines = 2*5 size with
    a nodal location (default).

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_matrix_field(3, 5, 2)

    """
    return _create_field(
        server,
        natures.matrix,
        num_entities,
        location,
        num_col,
        num_lines,
    )


def create_3d_vector_field(num_entities, location=locations.nodal, server=None):
    """Create a specific :class:`ansys.dpf.core.Field` with entities that have 3D vector format.

    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow the size of your field.

    Parameters
    ----------
    num_entities : int
        Number of entities to reserve

    location : str, optional
        Location of the field. Options are in :class:`locations <ansys.dpf.core.common.locations>`.
        The default is ``dpf.locations.nodal``.

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field of the requested format.

    Examples
    --------
    Create a field containing 4 3D vector entities with a nodal location (default).

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_3d_vector_field(4)

    """
    return _create_field(server, natures.vector, num_entities, location)


def create_tensor_field(num_entities, location=locations.nodal, server=None):
    """Create a specific :class:`ansys.dpf.core.Field` with entities that have a 3*3 format.

    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow the size of your field.

    Parameters
    ----------
    num_entities : int
        Number of entities to reserve.
    location : str, optional
        Location of the field. Options are in :class:`locations <ansys.dpf.core.common.locations>`.
        The default is ``dpf.locations.nodal``.

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field in the requested format.

    Examples
    --------
    Create a field containing 4 tensor entities with a nodal location (default).

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_tensor_field(4)

    """
    return _create_field(server, natures.symmatrix, num_entities, location)


def create_scalar_field(num_entities, location=locations.nodal, server=None):
    """Create a specific `:class:`ansys.dpf.core.Field` with entities that are scalar.

    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow the size of your field.

    Parameters
    ----------
    num_entities : int
        Number of entities to reserve
    location : str, optional
        Location of the field. Options are in :class:`locations <ansys.dpf.core.common.locations>`.
        The default is ``dpf.locations.nodal``.

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field in the requested format.

    Examples
    --------
    Create a field containing 4 scalars with a nodal location (default).

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_scalar_field(4)

    """
    return _create_field(server, natures.scalar, num_entities, location)


def create_vector_field(num_entities, num_comp, location=locations.nodal, server=None):
    """Create a specific `:class:`ansys.dpf.core.Field` with entities that have a vector format.

    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow the size of your field.

    Parameters
    ----------
    num_entities : int
        Number of entities to reserve.
    num_comp : int
        Number of vector components.
    location : str, optional
        Location of the field. Options are in :class:`locations <ansys.dpf.core.common.locations>`.
        The default is ``dpf.locations.nodal``.

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field in the requested format.

    Examples
    --------
    Create a field containing 3 vector entities of 5 components each with a
    nodal location (default).

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_vector_field(3, 5)

    """
    return _create_field(server, natures.vector, num_entities, location, ncomp_n=num_comp)


def create_overall_field(
    value, nature, num_entities, num_comp, location=locations.overall, server=None
):
    """Create a specific `:class:`ansys.dpf.core.Field` with entities that have an
    overall location.

    Regarding the nature of the entity contained in the field, we set the same value
    for all elements.

    Parameters
    ----------
    value : float
        Value of the entity
    nature : str
        Nature of the field entity data. For example:

        - :class:`ansys.dpf.core.natures.matrix`
        - :class:`ansys.dpf.core.natures.scalar`
    num_entities : int
        Number of entities to reserve.
    num_comp : int
        Number of vector components.
    location : str, optional
        Location of the field. Options are in :class:`locations <ansys.dpf.core.common.locations>`.
        The default is ``dpf.locations.nodal``.

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field in the requested format.

    Examples
    --------
    Create a field containing 10 scalar entities of 1 component each with an
    overall location (default). Same value (1.0) is set for all element of the field.

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_overall_field(1.0, natures.scalar, 10, 1)

    """
    overall_field = _create_field(server, nature, num_entities, location, ncomp_n=num_comp)
    for i in range(num_entities):
        overall_field.append(value, i)
    return overall_field


def _create_field(server, nature, nentities, location=locations.nodal, ncomp_n=0, ncomp_m=0):
    """Create a specific :class:`ansys.dpf.core.Field`.

    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow the size of your field.

    Parameters
    ----------
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.
    nature : str
        Nature of the field entity data. For example:

        - :class:`ansys.dpf.core.natures.matrix`
        - :class:`ansys.dpf.core.natures.scalar`

    nentities : int
        Number of entities to reserve.

    location : str, optional
        Location of the field. Options are in :class:`locations <ansys.dpf.core.common.locations>`.
        The default is ``dpf.locations.nodal``.

    ncomp_n : int
        Number of lines.
    ncomp_m : int
        Number of columns.

    Returns
    -------
    field : Field
        DPF field in the requested format.
    """
    if server is None:
        server = server_module.get_or_create_server(server)
    api = server.get_api_for_type(capi=field_capi.FieldCAPI, grpcapi=field_grpcapi.FieldGRPCAPI)
    api.init_field_environment(server)
    internal_obj = Field._field_create_internal_obj(
        api=api,
        client=server.client,
        nature=nature,
        nentities=nentities,
        location=location,
        ncomp_n=ncomp_n,
        ncomp_m=ncomp_m,
    )
    field = Field(field=internal_obj, server=server)
    return field



"""
average_over_label_fc
=====================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class average_over_label_fc(Operator):
    """Compute the component-wise average over all the fields that have the
    same ID as the label set as input in the fields container. This
    computation can be incremental. If the input fields container is
    connected and the operator is run multiple times, the output field
    will be on all the connected inputs.

    Parameters
    ----------
    fields_container : FieldsContainer
    label : str, optional
        Label of the fields container where it should
        operate.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.math.average_over_label_fc()

    >>> # Make input connections
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_label = str()
    >>> op.inputs.label.connect(my_label)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.math.average_over_label_fc(
    ...     fields_container=my_fields_container,
    ...     label=my_label,
    ... )

    >>> # Get output data
    >>> result_field = op.outputs.field()
    """

    def __init__(self, fields_container=None, label=None, config=None, server=None):
        super().__init__(name="average_over_label_fc", config=config, server=server)
        self._inputs = InputsAverageOverLabelFc(self)
        self._outputs = OutputsAverageOverLabelFc(self)
        if fields_container is not None:
            self.inputs.fields_container.connect(fields_container)
        if label is not None:
            self.inputs.label.connect(label)

    @staticmethod
    def _spec():
        description = """Compute the component-wise average over all the fields that have the
            same ID as the label set as input in the fields container.
            This computation can be incremental. If the input fields
            container is connected and the operator is run multiple
            times, the output field will be on all the connected
            inputs."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="label",
                    type_names=["string"],
                    optional=True,
                    document="""Label of the fields container where it should
        operate.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="field",
                    type_names=["field"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="average_over_label_fc", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsAverageOverLabelFc
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsAverageOverLabelFc
        """
        return super().outputs


class InputsAverageOverLabelFc(_Inputs):
    """Intermediate class used to connect user inputs to
    average_over_label_fc operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.math.average_over_label_fc()
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_label = str()
    >>> op.inputs.label.connect(my_label)
    """

    def __init__(self, op: Operator):
        super().__init__(average_over_label_fc._spec().inputs, op)
        self._fields_container = Input(
            average_over_label_fc._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._fields_container)
        self._label = Input(average_over_label_fc._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._label)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator.

        Parameters
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.average_over_label_fc()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> # or
        >>> op.inputs.fields_container(my_fields_container)
        """
        return self._fields_container

    @property
    def label(self):
        """Allows to connect label input to the operator.

        Label of the fields container where it should
        operate.

        Parameters
        ----------
        my_label : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.average_over_label_fc()
        >>> op.inputs.label.connect(my_label)
        >>> # or
        >>> op.inputs.label(my_label)
        """
        return self._label


class OutputsAverageOverLabelFc(_Outputs):
    """Intermediate class used to get outputs from
    average_over_label_fc operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.math.average_over_label_fc()
    >>> # Connect inputs : op.inputs. ...
    >>> result_field = op.outputs.field()
    """

    def __init__(self, op: Operator):
        super().__init__(average_over_label_fc._spec().outputs, op)
        self._field = Output(average_over_label_fc._spec().output_pin(0), 0, op)
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator

        Returns
        ----------
        my_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.math.average_over_label_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field()
        """  # noqa: E501
        return self._field



"""
change_cs
=========
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.outputs import _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class change_cs(Operator):
    """Applies a transformation (rotation and displacement) matrix on a mesh
    or meshes container.

    Parameters
    ----------
    meshes : MeshedRegion or MeshesContainer
    coordinate_system : Field
        3-3 rotation matrix + 3 translations (x, y,
        z)


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.mesh.change_cs()

    >>> # Make input connections
    >>> my_meshes = dpf.MeshedRegion()
    >>> op.inputs.meshes.connect(my_meshes)
    >>> my_coordinate_system = dpf.Field()
    >>> op.inputs.coordinate_system.connect(my_coordinate_system)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.mesh.change_cs(
    ...     meshes=my_meshes,
    ...     coordinate_system=my_coordinate_system,
    ... )

    >>> # Get output data
    >>> result_meshed_region = op.outputs.meshed_region()
    """

    def __init__(self, meshes=None, coordinate_system=None, config=None, server=None):
        super().__init__(name="mesh::change_cs", config=config, server=server)
        self._inputs = InputsChangeCs(self)
        self._outputs = OutputsChangeCs(self)
        if meshes is not None:
            self.inputs.meshes.connect(meshes)
        if coordinate_system is not None:
            self.inputs.coordinate_system.connect(coordinate_system)

    @staticmethod
    def _spec():
        description = """Applies a transformation (rotation and displacement) matrix on a mesh
            or meshes container."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="meshes",
                    type_names=["meshed_region", "meshes_container"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="coordinate_system",
                    type_names=["field"],
                    optional=False,
                    document="""3-3 rotation matrix + 3 translations (x, y,
        z)""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="meshed_region",
                    type_names=["meshed_region", "meshes_container"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="mesh::change_cs", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsChangeCs
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsChangeCs
        """
        return super().outputs


class InputsChangeCs(_Inputs):
    """Intermediate class used to connect user inputs to
    change_cs operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mesh.change_cs()
    >>> my_meshes = dpf.MeshedRegion()
    >>> op.inputs.meshes.connect(my_meshes)
    >>> my_coordinate_system = dpf.Field()
    >>> op.inputs.coordinate_system.connect(my_coordinate_system)
    """

    def __init__(self, op: Operator):
        super().__init__(change_cs._spec().inputs, op)
        self._meshes = Input(change_cs._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._meshes)
        self._coordinate_system = Input(change_cs._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._coordinate_system)

    @property
    def meshes(self):
        """Allows to connect meshes input to the operator.

        Parameters
        ----------
        my_meshes : MeshedRegion or MeshesContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.change_cs()
        >>> op.inputs.meshes.connect(my_meshes)
        >>> # or
        >>> op.inputs.meshes(my_meshes)
        """
        return self._meshes

    @property
    def coordinate_system(self):
        """Allows to connect coordinate_system input to the operator.

        3-3 rotation matrix + 3 translations (x, y,
        z)

        Parameters
        ----------
        my_coordinate_system : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mesh.change_cs()
        >>> op.inputs.coordinate_system.connect(my_coordinate_system)
        >>> # or
        >>> op.inputs.coordinate_system(my_coordinate_system)
        """
        return self._coordinate_system


class OutputsChangeCs(_Outputs):
    """Intermediate class used to get outputs from
    change_cs operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mesh.change_cs()
    >>> # Connect inputs : op.inputs. ...
    >>> result_meshed_region = op.outputs.meshed_region()
    """

    def __init__(self, op: Operator):
        super().__init__(change_cs._spec().outputs, op)
        self.meshed_region_as_meshed_region = Output(
            _modify_output_spec_with_one_type(
                change_cs._spec().output_pin(0), "meshed_region"
            ),
            0,
            op,
        )
        self._outputs.append(self.meshed_region_as_meshed_region)
        self.meshed_region_as_meshes_container = Output(
            _modify_output_spec_with_one_type(
                change_cs._spec().output_pin(0), "meshes_container"
            ),
            0,
            op,
        )
        self._outputs.append(self.meshed_region_as_meshes_container)
